Package: DiceView
Title: Methods for Visualization of Computer Experiments Design and
        Surrogate
Version: 3.2
Date: 2026-05-18
Authors@R: 
    c(person(given = "Yann",
             family = "Richet",
             role = c("aut","cre"),
             email = "yann.richet@asnr.fr",
             comment = c(ORCID = "0000-0002-5677-8458")),
      person(given = "Yves",
             family = "Deville",
             role = c("aut"),
             email = "deville.yves@alpestat.com"),
      person(given = "Clement",
             family = "Chevalier",
             role = c("ctb"),
             email = "clement.chevalier1@gmail.com"))
Maintainer: Yann Richet <yann.richet@asnr.fr>
Description: View 2D/3D sections, contour plots, mesh of excursion sets for computer experiments designs, surrogates or test functions.
Depends: methods, utils, stats, grDevices, graphics
Imports: DiceDesign, R.cache, geometry, scatterplot3d, parallel,
        foreach
Suggests: rlibkriging, DiceKriging, DiceEval, rgl, arrangements
License: GPL-3
URL: https://github.com/IRSN/DiceView
Repository: CRAN
RoxygenNote: 7.3.3
Encoding: UTF-8
NeedsCompilation: no
Packaged: 2026-05-18 18:30:13 UTC; richet
Author: Yann Richet [aut, cre] (ORCID: <https://orcid.org/0000-0002-5677-8458>),
  Yves Deville [aut],
  Clement Chevalier [ctb]
Date/Publication: 2026-05-19 05:40:02 UTC
Built: R 4.6.0; ; 2026-05-19 11:17:54 UTC; unix

# Generated by roxygen2: do not edit by hand

S3method(contourview,"function")
S3method(contourview,Kriging)
S3method(contourview,WarpKriging)
S3method(contourview,character)
S3method(contourview,glm)
S3method(contourview,km)
S3method(contourview,list)
S3method(contourview,matrix)
S3method(filledcontourview,"function")
S3method(filledcontourview,Kriging)
S3method(filledcontourview,WarpKriging)
S3method(filledcontourview,glm)
S3method(filledcontourview,km)
S3method(filledcontourview,list)
S3method(sectionview,"function")
S3method(sectionview,Kriging)
S3method(sectionview,WarpKriging)
S3method(sectionview,character)
S3method(sectionview,glm)
S3method(sectionview,km)
S3method(sectionview,list)
S3method(sectionview,matrix)
S3method(sectionview3d,"function")
S3method(sectionview3d,Kriging)
S3method(sectionview3d,WarpKriging)
S3method(sectionview3d,character)
S3method(sectionview3d,glm)
S3method(sectionview3d,km)
S3method(sectionview3d,list)
S3method(sectionview3d,matrix)
export(Apply.function)
export(Memoize.function)
export(Vectorize.function)
export(are_in.mesh)
export(branin)
export(contourview)
export(expand.grids)
export(filledcontourview)
export(is.mesh)
export(is_in.mesh)
export(is_in.p)
export(mesh)
export(mesh_exsets)
export(mesh_level)
export(min_dist)
export(min_dist.mesh)
export(optim.stop)
export(optims)
export(plot2d_mesh)
export(plot3d_mesh)
export(plot_mesh)
export(points_in.mesh)
export(points_out.mesh)
export(root)
export(roots)
export(roots_mesh)
export(sectionview)
export(sectionview3d)
import(grDevices)
import(graphics)
import(methods)
import(stats)
importFrom(DiceDesign,lhsDesign)
importFrom(R.cache,loadCache)
importFrom(R.cache,saveCache)
importFrom(foreach,"%dopar%")
importFrom(foreach,foreach)
importFrom(geometry,convhulln)
importFrom(geometry,delaunayn)
importFrom(geometry,inhulln)
importFrom(parallel,mclapply)
importFrom(utils,combn)
importFrom(utils,installed.packages)

[![R-CMD-check](https://github.com/IRSN/DiceView/actions/workflows/check-standard.yml/badge.svg)](https://github.com/IRSN/DiceView/actions/workflows/check-standard.yml)
[![submit-CRAN](https://github.com/IRSN/DiceView/actions/workflows/submit-on-cran.yml/badge.svg)](https://github.com/IRSN/DiceView/actions/workflows/submit-on-cran.yml)
[![CRAN-status](https://www.r-pkg.org/badges/version/DiceView)](https://cran.r-project.org/package=DiceView)
[![Downloads(monthly)](https://cranlogs.r-pkg.org/badges/DiceView)](https://cran.r-project.org/package=DiceView)

# DiceView: Methods for visualization of Computer Experiments Design and Surrogate

View 2D/3D sections, contour plots, mesh of excursion sets for computer experiments designs, surrogates or test functions.

Installation
------------

You can install the stabe version of the package using CRAN:

```
install.packages('DiceView')
```

Or install the latest version of the package using the `remotes` R package:

```
install.packages('remotes')
remotes::install_github("IRSN/DiceView")
```

![Analytics](https://ga-beacon.appspot.com/UA-109580-20/DiceView)

#' @param fun a function or 'predict()'-like function that returns a simple numeric or mean and standard error: list(mean=...,se=...).
#' @param vectorized is fun vectorized?
#' @param col_fading_interval an optional factor of alpha (color channel) fading used to plot function output intervals (if any).
#' @template contourview-doc
#' @rdname contourview
#' @method contourview function
#' @aliases contourview,function,function-method
#' @export
#' @seealso \code{\link{sectionview.function}} for a section plot, and \code{\link{sectionview3d.function}} for a 2D section plot.
#' @examples
#' x1 <- rnorm(15)
#' x2 <- rnorm(15)
#'
#' y <- x1 + x2 + rnorm(15)
#' model <- lm(y ~ x1 + x2)
#'
#' contourview(function(x) sum(x),
#'                      Xlim=cbind(range(x1),range(x2)), col='black')
#' points(x1,x2)
#'
#' contourview(function(x) {
#'                       x = as.data.frame(x)
#'                       colnames(x) <- all.vars(model$call)[-1]
#'                       predict.lm(model, newdata=x, se.fit=FALSE)
#'                     }, vectorized=TRUE, add=TRUE)
#'
contourview.function <- function(fun, vectorized=FALSE,
                             center = NULL,
                             lty_center = 2,
                             col_center = "black",
                             axis = NULL,
                             npoints = 21,
                             levels = 10,
                             lty_levels = 3,
                             col_levels = if (!is.null(col) & length(col)==1) col.levels(col,levels-1) else if (!is.null(col) & length(col)==2) cols.levels(col[1],col[2],levels-1) else col.levels("blue",levels-1),
                             col = NULL,
                             col_fading_interval = 0.5,
                             mfrow = NULL,
                             Xlab = NULL, ylab = NULL,
                             Xlim = if (!add) matrix(c(0,1),2,2) else NULL, ylim=NULL,
                             title = NULL, title_sep = " | ",
                             add = FALSE,
                             ...) {
    # setup for center/Xlim/axis/mfrow
    view = getView(2, add, center, Xlim, ylim, axis, mfrow)
    D = view$D
    center = view$center
    Xlim = view$Xlim
    ylim = view$ylim
    axis = view$axis
    mfrow = view$mfrow

    if (D == 1) stop("for a model with dim 1, use 'sectionview'")

    if (!add && (is.null(ylim) || all(is.na(ylim)))) { # compute and set ylim
        ylim <- range(unlist(EvalInterval.function(fun,Xlim, vectorized)), na.rm = TRUE)
        setView_ylim(ylim)
    }

    if ( (length(levels)==1) && is.wholenumber(levels)) {
        levels = pretty(range(unlist(EvalInterval.function(fun,X=Xlim,vectorized,dim=D)),na.rm=TRUE), levels)
        if (length(col_levels) != length(levels)) {
            warning(paste0("col_levels reset to ",length(levels)," values"))
            if (length(col_levels) == 2)
                col_levels = cols.levels(col_levels[1],col_levels[2],levels)
            else
                col_levels = col.levels(col_levels[1],levels)
        }
    }

    npoints <- rep(npoints, length.out = D)
    drx <- unlist(Xlim[2, ]) - unlist(Xlim[1, ])

    ## define X & y labels
    if (is.null(ylab)) ylab <- "y"
    if (is.null(Xlab)) Xlab <- paste(sep = "", "X", 1:D)

    ## try to find a good formatted value 'fcenter' for 'center'
    fcenter <- tryFormat(x = center, drx = drx)

    ## Each 'id' will produce a RGL plot
    for (id in 1:dim(axis)[1]) {
        if (D>2) screen(id, new=!add)

        d <- axis[id, ]

        npoints_all <- npoints[d[1]]*npoints[d[2]]

        ## ind.nonfix flags the non fixed dims
        ind.nonfix <- (1:D) %in% c(d[1], d[2])
        ind.nonfix <- !ind.nonfix

        x1lim <- Xlim[, d[1]]
        x2lim <- Xlim[, d[2]]

        xd1 <- seq(from = as.numeric(x1lim[1]), to = as.numeric(x1lim[2]), length.out = npoints[d[1]])
        xd2 <- seq(from = as.numeric(x2lim[1]), to = as.numeric(x2lim[2]), length.out = npoints[d[2]])

        x <- data.frame(t(matrix(as.numeric(center), nrow = D, ncol = npoints_all)))
        if (!is.null(center)) if(!is.null(names(center))) names(x) <- names(center)
        x[ , d] <- expand.grid(xd1, xd2)

        F_x = EvalInterval.function(fun, x, vectorized)
        F_x$yd <- matrix(F_x$y,ncol=npoints[2],nrow=npoints[1])
        F_x$yd_low <- matrix(F_x$y_low,ncol=npoints[2],nrow=npoints[1])
        F_x$yd_up <- matrix(F_x$y_up,ncol=npoints[2],nrow=npoints[1])

        if (is.null(title)){
            title_d <- paste(collapse = "~",sep = "~", ylab, paste(collapse = ",", sep = ",", Xlab[d[1]], Xlab[d[2]]))
            if (D>2) {
                title_d <-  paste(collapse =title_sep, sep =title_sep, title_d, paste(collapse=',',Xlab[ind.nonfix],'=', fcenter[ind.nonfix]))
            }
        } else {
            title_d <- title
        }

        ## plot mean
        if (isTRUE(add)) {
            contour(x = xd1,y = xd2, z = if (!all(is.na(F_x$yd))) F_x$yd else F_x$yd_low,
                    xlim = x1lim, ylim = x2lim, zlim = ylim,
                    col = col_levels,  lty = if (!all(is.na(F_x$yd))) lty_levels else 0,
                    levels = levels,
                    add=TRUE,
                    ...)
        } else {
            contour(x = xd1, y = xd2, z = if (!all(is.na(F_x$yd))) F_x$yd else F_x$yd_low,
                    xlab = Xlab[d[1]], ylab = Xlab[d[2]],
                    xlim = x1lim, ylim = x2lim, zlim = ylim,
                    main = title_d,
                    col = col_levels,  lty = if (!all(is.na(F_x$yd))) lty_levels else 0,
                    levels = levels,
                    add=FALSE,
                    ...)
            if(D>2) {
                abline(v=center[d[1]],col=col_center,lty=lty_center)
                abline(h=center[d[2]],col=col_center,lty=lty_center)
            }
        }

        ## 'confidence band' filled with the suitable color
	    if (!all(is.na(F_x$yd_low)) && !all(is.na(F_x$yd_up))) {
                for (i in 1:length(levels)) {
                    .filled.contour(x = xd1, y = xd2, z = abs((F_x$yd_low+F_x$yd_up)/2 - levels[i])-(F_x$yd_up-F_x$yd_low)/2,
                                    #xlim = xlim, ylim = ylim, zlim = zlim,
                                    col = fade(col_levels[i],alpha = 1-col_fading_interval),
                                    levels = c(-max(F_x$yd_up-F_x$yd_low),0))
            }
	    }
    }
}


#' @param X the matrix of input design.
#' @param y the array of output values (two columns means an interval).
#' @param col_points color of points.
#' @param bg_fading  an optional factor of alpha (color channel) fading used to plot design points outside from this section.
#' @template contourview-doc
#' @rdname contourview
#' @method contourview matrix
#' @aliases contourview,matrix,matrix-method
#' @export
#' @seealso \code{\link{sectionview.matrix}} for a section plot, and \code{\link{sectionview3d.matrix}} for a 2D section plot.
#' @examples
#' X = matrix(runif(15*2),ncol=2)
#' y = apply(X,1,branin)
#'
#' contourview(X, y)
#'
contourview.matrix <- function(X, y,
                           center = NULL,
                           lty_center = 2,
                           col_center = "black",
                           axis = NULL,
                           col_points = if (!is.null(col)) col else "red",
                           col = NULL,
                           bg_fading = 1,
                           mfrow = NULL,
                           Xlab = NULL, ylab = NULL,
                           Xlim = if (!add) matrix(c(0,1),2,2) else NULL, ylim=NULL,
                           title = NULL, title_sep = " | ",
                           add = FALSE,
                           ...) {
    # setup for center/Xlim/axis/mfrow
    view = getView(2, add, center, if (is.null(Xlim) && !add) apply(X, 2, range) else Xlim, if (is.null(ylim) && !add) ylim <- range(y) else ylim, axis, mfrow)
    D = view$D
    center = view$center
    Xlim = view$Xlim
    ylim = view$ylim
    axis = view$axis
    mfrow = view$mfrow

    if (ncol(X) != D)
        stop(paste0("X must have ",D," columns."))
    n <- nrow(X)

    if (is.matrix(y) && ncol(y) == 2) {
        y_low <- y[ , 1]
        y_up <- y[ , 2]
        y <- NA
    } else {
        y_low <- NA
        y_up <- NA
    }

    drx <- unlist(Xlim[2, ]) - unlist(Xlim[1, ])

    ## define X & y labels
    if (is.null(ylab) && !is.null(names(y))) ylab <- names(y)[1]
    if (is.null(Xlab)) Xlab <- names(X)

    ## define X & y labels
    if (is.null(ylab)) ylab <- "y"
    if (is.null(Xlab)) Xlab <- paste(sep = "", "X", 1:D)

    fcenter <- tryFormat(x = center, drx = drx)

    ## Each 'id' will produce a plot
    for (id in 1:dim(axis)[1]) {
        if (D>2) screen(id, new=!add)

        d <- axis[id,]

        ## ind.nonfix flags the non fixed dims
        ind.nonfix <- (1:D) %in% c(d[1], d[2])
        ind.nonfix <- !ind.nonfix

        x1lim = Xlim[, d[1]]
        x2lim = Xlim[, d[2]]

        if (is.null(title)){
            title_d <- paste(collapse = "~",sep = "~", ylab, paste(collapse = ",", sep = ",", Xlab[d[1]], Xlab[d[2]]))
            if (D>2) {
                title_d <-  paste(collapse =title_sep, sep =title_sep, title_d, paste(collapse=',',Xlab[ind.nonfix],'=', fcenter[ind.nonfix]))
            }
        } else {
            title_d <- title
        }

        ## fading colors for points
        if (D>2) {
            xrel <- scale(x = as.matrix(X),
                          center = center,
                          scale = drx)

            ## ind.nonfix flags the non fixed dims
            ind.nonfix <- (1:D) %in% c(d[1], d[2])
            ind.nonfix <- !ind.nonfix

            alpha <- pmax(0,apply(X = xrel[ , ind.nonfix, drop = FALSE],
                           MARGIN = 1,
                           FUN = function(x) (1 - sqrt(sum(x^2)/D))^bg_fading))
        } else {
            alpha <- rep(1, n)
        }

        if (isTRUE(add)) {
            if (D>2)
                plot(x=X[,d[1]], y=X[,d[2]],  # Cannot use 'points' so use 'plot' with these neutral args
                     col = fade(color = col_points, alpha = alpha),
                     pch = 20,type='p',
                     xlab="",ylab="", main="", xlim=x1lim, ylim=x2lim,
                     bty='n', xaxt='n', yaxt='n', ann=FALSE, # remove all text, that should be already     displayed
                     )
            else
                points(x=X[,d[1]], y=X[,d[2]],
                       col = fade(color = col_points, alpha = alpha),
                       xlim = x1lim, ylim = x2lim,
                       pch = 20)
        } else {
            plot(x=X[,d[1]], y=X[,d[2]],
                 xlab=Xlab[d[1]], ylab=Xlab[d[1]],
                 xlim=x1lim, ylim=x2lim,
                 main=title_d,
                 pch = 20, type = "p",
                 col = fade(color = col_points, alpha = alpha),
                 ...)
            if(D>2) {
                abline(v=center[d[1]],col=col_center,lty=lty_center)
                abline(h=center[d[2]],col=col_center,lty=lty_center)
            }
        }
    }
}

#' @param eval_str the expression to evaluate in each subplot.
#' @param axis optional matrix of 2-axis combinations to plot, one by row. The value \code{NULL} leads to all possible combinations i.e. \code{choose(D, 2)}.
#' @param mfrow  an optional list to force \code{par(mfrow = ...)} call. The default value  \code{NULL} is automatically set for compact view.
#' @rdname contourview
#' @method contourview character
#' @aliases contourview,character,character-method
#' @export
#' @seealso \code{\link{contourview.matrix}} for a section plot.
#' @examples
#' x1 <- rnorm(15)
#' x2 <- rnorm(15)
#'
#' y <- x1 + x2^2 + rnorm(15)
#' model <- glm(y ~ x1 + I(x2^2))
#'
#' contourview(model)
#'
#' contourview("abline(h=0.25,col='red')")
contourview.character <- function(eval_str,
                                 axis = NULL,
                                 mfrow = NULL,
                                 ...) {

    if (!exists(".split.screen.lim",envir=DiceView.env))
        stop(paste0("Cannot eval '",eval_str,"' when no previous sectionview() was called."))

    # setup for center/Xlim/axis/mfrow
    view = getView(2, TRUE, NA, NA, NA, axis, mfrow)
    D = view$D
    center = view$center
    Xlim = view$Xlim
    ylim = view$ylim
    axis = view$axis
    mfrow = view$mfrow

    ## Each 'id' will produce a plot
    for (id in 1:dim(axis)[1]) {

        d <- axis[id,]

        e = parent.frame()
        assign("d",d,envir=e)
        assign("xlim",Xlim[,d[1]],envir=e)
        assign("ylim",Xlim[,d[2]],envir=e)
        assign("zlim",ylim,envir=e)

        if (D>2) {
            screen(id, new=FALSE)
            plot(x=Xlim[,d[1]], y=Xlim[,d[2]],
                 type='n',
                 xlab="",ylab="", main="",
                 bty='n', xaxt='n', yaxt='n', ann=FALSE, # remove all text, that should be already displayed
                 ...)
        }

        eval(parse(text = eval_str), envir = e)
    }
}

#' @param km_model an object of class \code{"km"}.
#' @param type the kriging type to use for model prediction.
#' @param col_points color of points.
#' @param conf_level confidence hulls to display.
#' @param conf_fading an optional factor of alpha (color channel) fading used to plot confidence intervals.
#' @param bg_fading  an optional factor of alpha (color channel) fading used to plot design points outside from this section.
#' @template contourview-doc
#' @rdname contourview
#' @method contourview km
#' @aliases contourview,km,km-method
#' @export
#' @seealso \code{\link{sectionview.km}} for a section plot, and \code{\link{sectionview3d.km}} for a 2D section plot.
#' @examples
#' if (requireNamespace("DiceKriging")) { library(DiceKriging)
#'
#' X = matrix(runif(15*2),ncol=2)
#' y = apply(X,1,branin)
#'
#' model <- km(design = X, response = y, covtype="matern3_2")
#'
#' contourview(model)
#'
#' }
#'
contourview.km <- function(km_model, type = "UK",
                           center = NULL,
                           axis = NULL,
                           npoints = 21,
                           levels = pretty(km_model@y, 10),
                           col_points = if (!is.null(col) & length(col)==1) col else "red",
                           col_levels = if (!is.null(col) & length(col)==1) col.levels(col,levels) else if (!is.null(col) & length(col)==2) cols.levels(col[1],col[2],levels-1) else col.levels("blue",levels),
                           col = NULL,
                           conf_level = 0.5,
                           conf_fading = 0.5,
                           bg_fading = 1,
                           mfrow = NULL,
                           Xlab = NULL, ylab = NULL,
                           Xlim = NULL, ylim=NULL,
                           title = NULL, title_sep = " | ",
                           add = FALSE,
                           ...) {
    if ( (length(levels)==1) && is.wholenumber(levels)) {
        levels = pretty(c(km_model@y+2*sqrt(km_model@covariance@sd2), km_model@y-2*sqrt(km_model@covariance@sd2)), levels)
        if (length(col_levels) != length(levels)) {
            warning(paste0("col_levels reset to ",length(levels)," values"))
            if (length(col_levels) == 2)
                col_levels = cols.levels(col_levels[1],col_levels[2],levels)
            else
                col_levels = col.levels(col_levels[1],levels)
        }
    }

    X_doe <- km_model@X
    y_doe <- km_model@y

    D <- ncol(X_doe)
    n <- nrow(X_doe)

    if (km_model@noise.flag) {
        sdy_doe <- sqrt(km_model@noise.var)
    } else if (km_model@covariance@nugget.flag) {
        sdy_doe <- rep(sqrt(km_model@covariance@nugget), n)
    } else {
        sdy_doe <- 0
    }

    ## find limits
    if (is.null(Xlim) && !add)
        Xlim <- apply(X_doe, 2, range)
    else if (!is.null(Xlim))
        Xlim <- matrix(Xlim,nrow=2,ncol=D)

    if (is.null(ylim) && !add) {
        ymin <- min(y_doe-3*sdy_doe)
        ymax <- max(y_doe+3*sdy_doe)
        ylim <- c(ymin, ymax)
    }

    ## define X & y labels
    if (is.null(ylab)) ylab <- names(y_doe)
    if (is.null(Xlab)) Xlab <- names(X_doe)

    if (is.null(conf_fading) ||
        length(conf_fading) != length(conf_level))
        conf_fading <- rep(0.5/length(conf_level), length(conf_level))

    # plot mean
    contourview.function(fun = function(x) {
            DiceKriging::predict.km(km_model,type=type,newdata=x,checkNames=FALSE)$mean
        }, vectorized=TRUE,
        center = center, axis = axis, mfrow = mfrow, Xlim = Xlim, ylim=ylim,
        npoints = npoints,
        levels = levels, col_levels = col_levels,
        Xlab = Xlab, ylab = ylab,
        title = title, title_sep  = title_sep, add = add, ...)

    # plot confidence bands
    for (l in conf_level) {
        contourview.function(fun = function(x) {
                p = DiceKriging::predict.km(km_model,type=type,newdata=x,checkNames=FALSE)
                cbind(p$mean-qnorm(1-(1-l)/2) * p$sd, p$mean+qnorm(1-(1-l)/2) * p$sd)
            }, vectorized=TRUE,
            npoints = npoints,
            levels = levels, col_levels = col_levels,
            col_fading_interval=conf_fading,
            add = TRUE)
    }

    # plot design points
    contourview.matrix(X = X_doe, y = y_doe,
                       col_points = col_points,
                       bg_fading = bg_fading,
                       add=TRUE)
}

#' @param libKriging_model an object of class \code{"Kriging"} or \code{"WarpKriging"}.
#' @param col_points color of points.
#' @param conf_level confidence level hull to display.
#' @param col_fading_interval an optional factor of alpha (color channel) fading used to plot confidence hull.
#' @param bg_fading  an optional factor of alpha (color channel) fading used to plot design points outside from this section.
contourview_libKriging <- function(libKriging_model,
                           center = NULL,
                           axis = NULL,
                           npoints = 21,
                           levels = pretty( libKriging_model$y() , 10),
                           col_points = if (!is.null(col) & length(col)==1) col else "red",
                           col_levels = if (!is.null(col) & length(col)==1) col.levels(col,levels) else if (!is.null(col) & length(col)==2) cols.levels(col[1],col[2],levels-1) else col.levels("blue",levels),
                           col = NULL,
                           conf_level = 0.5,
                           conf_fading = 0.5,
                           bg_fading = 1,
                           mfrow = NULL,
                           Xlab = NULL, ylab = NULL,
                           Xlim = NULL, ylim=NULL,
                           title = NULL, title_sep = " | ",
                           add = FALSE,
                           ...) {
    if ( (length(levels)==1) && is.wholenumber(levels)) {
        levels = pretty(c(libKriging_model$y()-2*sqrt(libKriging_model$sigma2()),
                          libKriging_model$y()+2*sqrt(libKriging_model$sigma2())), levels)
        if (length(col_levels) != length(levels)) {
            warning(paste0("col_levels reset to ",length(levels)," values"))
            if (length(col_levels) == 2)
                col_levels = cols.levels(col_levels[1],col_levels[2],levels)
            else
                col_levels = col.levels(col_levels[1],levels)
        }
    }

    X_doe <- libKriging_model$X()
    y_doe <- libKriging_model$y()

    D <- ncol(X_doe)
    n <- nrow(X_doe)

    if (inherits(libKriging_model, "Kriging")) {
        nm <- libKriging_model$noise_model()
        if (nm == "nugget") {
            sdy_doe <- rep(sqrt(libKriging_model$nugget()), n)
        } else if (nm == "heterogeneous") {
            sdy_doe <- sqrt(libKriging_model$noise())
        } else {
            sdy_doe <- rep(0, n)
        }
    } else if (inherits(libKriging_model, "WarpKriging")) {
        sdy_doe <- rep(0, n)
    } else
        stop(paste0("Kriging model of class ",class(libKriging_model)," is not yet supported."))

    ## find limits
    if (is.null(Xlim) && !add)
        Xlim <- apply(X_doe, 2, range)
    else if (!is.null(Xlim))
        Xlim <- matrix(Xlim,nrow=2,ncol=D)

    if (is.null(ylim) && !add) {
        if (is.null(sdy_doe))
            ylim = range(y_doe)
        else {
            ymin <- min(y_doe-3*sdy_doe)
            ymax <- max(y_doe+3*sdy_doe)
            ylim <- c(ymin, ymax)
        }
    }

    ## define X & y labels
    if (is.null(ylab)) ylab <- names(y_doe)
    if (is.null(ylab)) ylab <- "y"
    if (is.null(Xlab)) Xlab <- names(X_doe)
    if (is.null(Xlab)) Xlab <- paste(sep = "", "X", 1:D)

    if (is.null(conf_fading) ||
        length(conf_fading) != length(conf_level))
        conf_fading <- rep(0.5/length(conf_level), length(conf_level))

    # plot mean
    contourview.function(fun = function(x) {
            rlibkriging::predict(libKriging_model,x,return_stdev=FALSE)$mean
        }, vectorized=TRUE,
        center = center, axis = axis, mfrow = mfrow, Xlim = Xlim, ylim=ylim,
        npoints = npoints,
        levels = levels, col_levels = col_levels,
        Xlab = Xlab, ylab = ylab,
        title = title, title_sep = title_sep, add = add, ...)

    # plot confidence bands
    for (l in conf_level) {
        contourview.function(fun = function(x) {
                p = rlibkriging::predict(libKriging_model,x,return_stdev=TRUE)
                cbind(p$mean-qnorm(1-(1-l)/2) * p$stdev, p$mean+qnorm(1-(1-l)/2) * p$stdev)
            }, vectorized=TRUE,
            npoints = npoints,
            levels = levels, col_levels = col_levels,
            col_fading_interval=conf_fading,
            add = TRUE)
    }

    # plot design points
    contourview.matrix(X = X_doe, y = y_doe,
                       col_points = col_points,
                       bg_fading = bg_fading,
                       add=TRUE)
}

#' @param Kriging_model an object of class \code{"Kriging"}.
#' @param col_points color of points.
#' @param conf_level confidence hulls to display.
#' @param conf_fading an optional factor of alpha (color channel) fading used to plot confidence intervals.
#' @param bg_fading  an optional factor of alpha (color channel) fading used to plot design points outside from this section.
#' @template contourview-doc
#' @rdname contourview
#' @method contourview Kriging
#' @aliases contourview,Kriging,Kriging-method
#' @export
#' @seealso \code{\link{sectionview.Kriging}} for a section plot, and \code{\link{sectionview3d.Kriging}} for a 2D section plot.
#' @examples
#' if (requireNamespace("rlibkriging")) { library(rlibkriging)
#'
#' X = matrix(runif(15*2),ncol=2)
#' y = apply(X,1,branin)
#'
#' model <- Kriging(X = X, y = y, kernel="matern3_2")
#'
#' contourview(model)
#'
#' }
#'
contourview.Kriging <- function(Kriging_model,
                                   center = NULL,
                                   axis = NULL,
                                   npoints = 21,
                                   levels = pretty( Kriging_model$y() , 10),
                                   col_points = if (!is.null(col) & length(col)==1) col else "red",
                                   col_levels = if (!is.null(col) & length(col)==1) col.levels(col,levels) else if (!is.null(col) & length(col)==2) cols.levels(col[1],col[2],levels-1) else col.levels("blue",levels),
                                   col = NULL,
                                   conf_level = 0.5,
                                   conf_fading = 0.5,
                                   bg_fading = 1,
                                   mfrow = NULL,
                                   Xlab = NULL, ylab = NULL,
                                   Xlim = NULL, ylim=NULL,
                                   title = NULL, title_sep = " | ",
                                   add = FALSE,
                                   ...) {
    contourview_libKriging(libKriging_model = Kriging_model,
                           center = center,
                           axis = axis,
                           npoints = npoints,
                           levels = levels,
                           col_points = col_points,
                           col_levels = col_levels,
                           col = col,
                           conf_level = conf_level,
                           conf_fading = conf_fading,
                           bg_fading = bg_fading,
                           mfrow = mfrow,
                           Xlab = Xlab, ylab = ylab,
                           Xlim = Xlim, ylim=ylim,
                           title = title,
                           add = add,
                           ...)
}

#' @param WarpKriging_model an object of class \code{"WarpKriging"}.
#' @param col_points color of points.
#' @param conf_level an optional list of confidence hulls to display.
#' @param conf_fading an optional factor of alpha (color channel) fading used to plot confidence intervals.
#' @param bg_fading  an optional factor of alpha (color channel) fading used to plot design points outside from this section.
#' @template contourview-doc
#' @rdname contourview
#' @method contourview WarpKriging
#' @aliases contourview,WarpKriging,WarpKriging-method
#' @export
#' @seealso \code{\link{sectionview.WarpKriging}} for a section plot, and \code{\link{sectionview3d.WarpKriging}} for a 2D section plot.
#' @examples
#' if (requireNamespace("rlibkriging")) { library(rlibkriging)
#'
#' X = matrix(runif(15*2),ncol=2)
#' y = apply(X,1,branin) + 5*rnorm(15)
#'
#' model <- WarpKriging(y = y, X = X, warping = c("affine","affine"), kernel="matern3_2")
#'
#' contourview(model)
#'
#' }
#'
contourview.WarpKriging <- function(WarpKriging_model,
                                center = NULL,
                                axis = NULL,
                                npoints = 21,
                                levels = pretty( WarpKriging_model$y() , 10),
                                col_points = if (!is.null(col) & length(col)==1) col else "red",
                                col_levels = if (!is.null(col) & length(col)==1) col.levels(col,levels) else if (!is.null(col) & length(col)==2) cols.levels(col[1],col[2],levels-1) else col.levels("blue",levels),
                                col = NULL,
                                conf_level = 0.5,
                                conf_fading = 0.5,
                                bg_fading = 1,
                                mfrow = NULL,
                                Xlab = NULL, ylab = NULL,
                                Xlim = NULL, ylim=NULL,
                                title = NULL, title_sep = " | ",
                                add = FALSE,
                                ...) {
    contourview_libKriging(libKriging_model = WarpKriging_model,
                           center = center,
                           axis = axis,
                           npoints = npoints,
                           levels = levels,
                           col_points = col_points,
                           col_levels = col_levels,
                           col = col,
                           conf_level = conf_level,
                           conf_fading = conf_fading,
                           bg_fading = bg_fading,
                           mfrow = mfrow,
                           Xlab = Xlab, ylab = ylab,
                           Xlim = Xlim, ylim=ylim,
                           title = title,
                           add = add,
                           ...)
    }

#' @param glm_model an object of class \code{"glm"}.
#' @param col_points color of points.
#' @param conf_level confidence hulls to display.
#' @param conf_fading an optional factor of alpha (color channel) fading used to plot confidence hull.
#' @param bg_fading  an optional factor of alpha (color channel) fading used to plot design points outside from this section.
#' @template contourview-doc
#' @rdname contourview
#' @method contourview glm
#' @aliases contourview,glm,glm-method
#' @export
#' @seealso \code{\link{sectionview.glm}} for a section plot, and \code{\link{sectionview3d.glm}} for a 2D section plot.
#' @examples
#' x1 <- rnorm(15)
#' x2 <- rnorm(15)
#'
#' y <- x1 + x2^2 + rnorm(15)
#' model <- glm(y ~ x1 + I(x2^2))
#'
#' contourview(model)
#'
contourview.glm <- function(glm_model,
                           center = NULL,
                           axis = NULL,
                           npoints = 21,
                           levels = pretty( glm_model$fitted.values , 10),
                           col_points = if (!is.null(col) & length(col)==1) col else "red",
                           col_levels = if (!is.null(col) & length(col)==1) col.levels(col,levels) else if (!is.null(col) & length(col)==2) cols.levels(col[1],col[2],levels-1) else col.levels("blue",levels),
                           col = NULL,
                           conf_level = 0.5,
                           conf_fading = 0.5,
                           bg_fading = 1,
                           mfrow = NULL,
                           Xlab = NULL, ylab = NULL,
                           Xlim = NULL, ylim=NULL,
                           title = NULL, title_sep = " | ",
                           add = FALSE,
                           ...) {
    if ( (length(levels)==1) && is.wholenumber(levels)) {
        levels = pretty(glm_model$fitted.values, levels)
        if (length(col_levels) != length(levels)) {
            warning(paste0("col_levels reset to ",length(levels)," values"))
            if (length(col_levels) == 2)
                col_levels = cols.levels(col_levels[1],col_levels[2],levels)
            else
                col_levels = col.levels(col_levels[1],levels)
        }
    }

    # Get X & y labels
    if (is.null(Xlab)) Xlab <- all.vars(glm_model$formula)[-1] # assume just one y
    if (is.null(ylab)) {
        factors_names = all.vars(glm_model$formula)
        for (n in Xlab) {
            factors_names = factors_names[-which(factors_names==n)]
        }
        ylab = factors_names
    }

    D <- length(Xlab)
    n <- length(glm_model$residuals)

    X_doe <- do.call(cbind,lapply(Xlab,function(Xn)glm_model$data[[Xn]]))
    colnames(X_doe) <- Xlab
    y_doe <- do.call(cbind,lapply(ylab,function(yn)glm_model$data[[yn]]))
    colnames(y_doe) <- ylab

    ## find limits
    if (is.null(Xlim) && !add)
        Xlim <- apply(X_doe, 2, range)
    else if (!is.null(Xlim))
        Xlim <- matrix(Xlim,nrow=2,ncol=D)

    if (is.null(conf_fading) ||
        length(conf_fading) != length(conf_level))
        conf_fading <- rep(0.5/length(conf_level), length(conf_level))

    # plot mean
    contourview.function(
        fun = function(x) {
            x = as.data.frame(x)
            colnames(x) <- Xlab
            predict.glm(glm_model, newdata=x, se.fit=FALSE)
        }, vectorized=TRUE,
        center = center,axis = axis, mfrow = mfrow, Xlim = Xlim, ylim=range(y_doe),
        npoints = npoints,
        levels = levels, col_levels = col_levels,
        Xlab = Xlab, ylab = ylab,
        title = title, title_sep  = title_sep, add = add, ...)

                           # plot confidence bands
    for (l in conf_level) {
        contourview.function(fun = function(x) {
                x = as.data.frame(x)
                colnames(x) <- Xlab
                p = predict.glm(glm_model, newdata=x, se.fit=TRUE)
                cbind(p$fit-qnorm(1-(1-l)/2) * p$se.fit, p$fit+qnorm(1-(1-l)/2) * p$se.fit)
            }, vectorized=TRUE,
            npoints = npoints,
            levels = levels, col_levels = col_levels,
            col_fading_interval=conf_fading,
            add = TRUE)
    }

    # plot design points
    contourview.matrix(X = X_doe, y = y_doe,
                       col_points = col_points,
                       bg_fading = bg_fading,
                       add=TRUE)
}


#' @param modelFit_model an object returned by DiceEval::modelFit.
#' @param col_points color of points.
#' @param bg_fading  an optional factor of alpha (color channel) fading used to plot design points outside from this section.
#' @template contourview-doc
#' @rdname contourview
#' @method contourview list
#' @aliases contourview,list,list-method
#' @export
#' @seealso \code{\link{sectionview.glm}} for a section plot, and \code{\link{sectionview3d.glm}} for a 2D section plot.
#' @examples
#' if (requireNamespace("DiceEval")) { library(DiceEval)
#'
#' X = matrix(runif(15*2),ncol=2)
#' y = apply(X,1,branin)
#'
#' model <- modelFit(X, y, type = "StepLinear")
#'
#' contourview(model)
#'
#' }
#'
contourview.list <- function(modelFit_model,
                            center = NULL,
                            axis = NULL,
                            npoints = 21,
                            levels = pretty( modelFit_model$data$Y , 10),
                            col_points = if (!is.null(col) & length(col)==1) col else "red",
                            col_levels = if (!is.null(col) & length(col)==1) col.levels(col,levels) else if (!is.null(col) & length(col)==2) cols.levels(col[1],col[2],levels-1) else col.levels("blue",levels),
                            col = NULL,
                            bg_fading = 1,
                            mfrow = NULL,
                            Xlab = NULL, ylab = NULL,
                            Xlim = NULL, ylim=NULL,
                            title = NULL, title_sep = " | ",
                            add = FALSE,
                            ...) {
    if ( (length(levels)==1) && is.wholenumber(levels)) {
        levels = pretty(modelFit_model$data$Y, levels)
        if (length(col_levels) != length(levels)) {
            warning(paste0("col_levels reset to ",length(levels)," values"))
            if (length(col_levels) == 2)
                col_levels = cols.levels(col_levels[1],col_levels[2],levels)
            else
                col_levels = col.levels(col_levels[1],levels)
        }
    }

    X_doe <- modelFit_model$data$X
    y_doe <- modelFit_model$data$Y

    D <- ncol(X_doe)
    n <- nrow(X_doe)

    ## define X & y labels
    if (is.null(ylab)) ylab <- names(y_doe)
    if (is.null(ylab)) ylab <- "y"
    if (is.null(Xlab)) Xlab <- names(X_doe)
    if (is.null(Xlab)) Xlab <- paste(sep = "", "X", 1:D)

    ## find limits
    if (is.null(Xlim) && !add)
        Xlim <- apply(X_doe, 2, range)
    else if (!is.null(Xlim))
        Xlim <- matrix(Xlim,nrow=2,ncol=D)

    # plot mean
    contourview.function(
        fun = function(x) {
            x = as.data.frame(x)
            colnames(x) <- Xlab
            DiceEval::modelPredict(modelFit_model, x)
        }, vectorized=TRUE,
        center = center,axis = axis,mfrow = mfrow,Xlim = Xlim, ylim=range(y_doe),
        npoints = npoints,
        levels = levels, col_levels = col_levels,
        Xlab = Xlab, ylab = ylab,
        title = title, title_sep  = title_sep, add = add, ...)

    # plot design points
    contourview.matrix(X = X_doe, y = y_doe,
                       col_points = col_points,
                       bg_fading = bg_fading,
                       add=TRUE)
}



#### Wrapper for contourview ####

#' @import methods
if(!isGeneric("contourview")) {
    setGeneric(name = "contourview",
               def = function(...) standardGeneric("contourview")
    )
}

#' @title Plot a contour view of a prediction model or function, including design points if available.
#' @details If available, experimental points are plotted with fading colors. Points that fall in the specified section (if any) have the color specified \code{col_points} while points far away from the center have shaded versions of the same color. The amount of fading is determined using the Euclidean distance between the plotted point and \code{center}.
#' @param ... arguments of the \code{contourview.km}, \code{contourview.glm}, \code{contourview.Kriging} or \code{contourview.function} function
#' @export
#' @examples
#' ## A 2D example - Branin-Hoo function
#' contourview(branin, levels=30, col='black')
#'
#' \dontrun{
#' ## a 16-points factorial design, and the corresponding response
#' d <- 2; n <- 16
#' design.fact <- expand.grid(seq(0, 1, length = 4), seq(0, 1, length = 4))
#' design.fact <- data.frame(design.fact); names(design.fact) <- c("x1", "x2")
#' y <- branin(design.fact); names(y) <- "y"
#'
#' if (requireNamespace("DiceKriging")) { library(DiceKriging)
#' ## model: km
#' model <- DiceKriging::km(design = design.fact, response = y)
#' contourview(model, levels=30)
#' contourview(branin, levels=30, col='red', add=TRUE)
#' }
#'
#' if (requireNamespace("rlibkriging")) { library(rlibkriging)
#' ## model: Kriging
#' model <- Kriging(X = as.matrix(design.fact), y = as.matrix(y), kernel="matern3_2")
#' contourview(model, levels=30)
#' contourview(branin, levels=30, col='red', add=TRUE)
#' }
#'
#' ## model: glm
#' model <- glm(y ~ 1+ x1 + x2 + I(x1^2) + I(x2^2) + x1*x2, data=cbind(y,design.fact))
#' contourview(model, levels=30)
#' contourview(branin, levels=30, col='red', add=TRUE)
#'
#' if (requireNamespace("DiceEval")) { library(DiceEval)
#' ## model: StepLinear
#' model <- modelFit(design.fact, y, type = "StepLinear")
#' contourview(model, levels=30)
#' contourview(branin, levels=30, col='red', add=TRUE)
#' }
#' }
#'
contourview <- function(...){
    UseMethod("contourview")
}

#' @param fun a function or 'predict()'-like function that returns a simple numeric, or an interval, or mean and standard error: list(mean=...,se=...).
#' @param vectorized is fun vectorized?
#' @param col_fun color of the function plot.
#' @param col_fading_interval an optional factor of alpha (color channel) fading used to plot function output intervals (if any).
#' @template sectionview-doc
#' @rdname sectionview
#' @method sectionview function
#' @aliases sectionview,function,function-method
#' @export
#' @seealso \code{\link{sectionview.function}} for a section plot, and \code{\link{sectionview3d.function}} for a 2D section plot.
#' @examples
#' x1 <- rnorm(15)
#' x2 <- rnorm(15)
#'
#' y <- x1 + x2 + rnorm(15)
#' model <- lm(y ~ x1 + x2)
#'
#' sectionview(function(x) sum(x),
#'                      center=c(0,0), Xlim=cbind(range(x1),range(x2)), col='black')
#'
#' sectionview(function(x) {
#'                       x = as.data.frame(x)
#'                       colnames(x) <- all.vars(model$call)[-1]
#'                       p = predict.lm(model, newdata=x, se.fit=TRUE)
#'                       cbind(p$fit-1.96 * p$se.fit, p$fit+1.96 * p$se.fit)
#'                     }, vectorized=TRUE, add=TRUE)
#'
sectionview.function <- function(fun, vectorized=FALSE,
                             center = NULL,
                             lty_center = 2,
                             col_center = "black",
                             axis = NULL,
                             npoints = 101,
                             col_fun = if (!is.null(col)) col else "blue",
                             col = NULL,
                             col_fading_interval = 0.5,
                             mfrow = NULL,
                             Xlab = NULL, ylab = NULL,
                             Xlim = if (!add) c(0,1) else NULL, ylim = NULL,
                             title = NULL, title_sep = " | ",
                             add = FALSE,
                             ...) {
    # setup for center/Xlim/axis/mfrow
    view = getView(1, add, center, Xlim, ylim, axis, mfrow)
    D = view$D
    center = view$center
    Xlim = view$Xlim
    ylim = view$ylim
    axis = view$axis
    mfrow = view$mfrow

    if (!add && (is.null(ylim) || all(is.na(ylim)))) { # compute and set ylim
        ylim <- range(unlist(EvalInterval.function(fun,Xlim, vectorized)), na.rm = TRUE)
        setView_ylim(ylim)
    }

    npoints <- rep(npoints, length.out = D)
    drx <- unlist(Xlim[2, ]) - unlist(Xlim[1, ])

    ## define X & y labels
    if (is.null(ylab)) ylab <- "y"
    if (is.null(Xlab)) Xlab <- paste(sep = "", "X", 1:D)

    ## try to find a good formatted value 'fcenter' for 'center'
    fcenter <- tryFormat(x = center, drx = drx)

    ## Each 'id' will produce a plot
    for (id in 1:dim(axis)[1]) {
        if (D>1) screen(id, new=!add)

        d <- axis[id, ]

        xlim = Xlim[, d]

        xd <- seq(from = xlim[1], to = xlim[2], length.out = npoints[d])

        x <- data.frame(t(matrix(as.numeric(center), nrow = D, ncol = npoints[d])))
        if (!is.null(center)) if(!is.null(names(center))) names(x) <- names(center)
        x[ , d] <- xd

        F_x = EvalInterval.function(fun, x, vectorized)

        if (is.null(title)){
            title_d <- paste(collapse = "~",sep = "~", ylab, Xlab[d])
            if (D>1) {
                title_d <-  paste(collapse =title_sep, sep =title_sep, title_d, paste(collapse=',',Xlab[-d], '=', fcenter[-d]))
            }
        } else {
            title_d <- title
        }

        ## plot mean
        if (isTRUE(add)) {
            if (D>1) {
                plot(xd, if (!all(is.na(F_x$y))) F_x$y else F_x$y_low,
                     xlim = xlim, ylim  =ylim,
                     type = if (!all(is.na(F_x$y))) "l" else "n",
                     col = col_fun, xlab="", ylab="",
                     bty='n', xaxt='n', yaxt='n', ann=FALSE,# remove all text, that should be already displayed
                     ...)
            } else { # not using screen(), so need for a non reset plotting method
                lines(xd, if (!all(is.na(F_x$y))) F_x$y else F_x$y_low,
                      xlim = xlim, ylim = ylim,
                      type = if (!all(is.na(F_x$y))) "l" else "n",
                      col = col_fun,
                      ...)
            }
        } else {
            plot(xd, if (!all(is.na(F_x$y))) F_x$y else F_x$y_low,
                 xlab = Xlab[d], ylab = ylab,
                 xlim = xlim, ylim = ylim,
                 main = title_d,
                 type = if (!all(is.na(F_x$y))) "l" else "n",
                 col = col_fun,
                 ...)
            if(D>1) abline(v=center[d],col=col_center,lty=lty_center)
        }

        ## 'confidence band' filled with the suitable color
	    if (!all(is.na(F_x$y_low)) && !all(is.na(F_x$y_up))) {
                polygon(c(xd,rev(xd)),
                        c(F_x$y_low,rev(F_x$y_up)),
                        col = translude(col_fun, alpha = col_fading_interval),
                        border = NA)
	    }
    }
}


#' @param X the matrix of input design.
#' @param y the array of output values (two columns means an interval).
#' @param col_points color of points.
#' @param col_fading_interval an optional factor of alpha (color channel) fading used to plot confidence intervals.
#' @param bg_fading  an optional factor of alpha (color channel) fading used to plot design points outside from this section.
#' @template sectionview-doc
#' @rdname sectionview
#' @method sectionview matrix
#' @aliases sectionview,matrix,matrix-method
#' @export
#' @seealso \code{\link{sectionview.matrix}} for a section plot, and \code{\link{sectionview3d.matrix}} for a 2D section plot.
#' @examples
#' X = matrix(runif(15*2),ncol=2)
#' y = apply(X,1,branin)
#'
#' sectionview(X,y, center=c(.5,.5))
#'
sectionview.matrix <- function(X, y,
                               center = NULL,
                               lty_center = 2,
                               col_center = "black",
                               axis = NULL,
                               col_points = if (!is.null(col)) col else "red",
                               col = NULL,
                               col_fading_interval = 0.5,
                               bg_fading = 5,
                               mfrow = NULL,
                               Xlab = NULL, ylab = NULL,
                               Xlim = if (!add) c(0,1) else NULL, ylim=NULL,
                               title = NULL, title_sep = " | ",
                               add = FALSE,
                               ...) {
    # setup for center/Xlim/axis/mfrow
    view = getView(1, add, center, if (is.null(Xlim) && !add) apply(X, 2, range) else Xlim, if (is.null(ylim) && !add) ylim <- range(y) else ylim, axis, mfrow)
    D = view$D
    center = view$center
    Xlim = view$Xlim
    ylim = view$ylim
    axis = view$axis
    mfrow = view$mfrow

    if (ncol(X) != D)
        stop(paste0("X must have ",D," columns."))
    n <- nrow(X)

    if (is.matrix(y) && ncol(y) == 2) {
        y_low <- y[ , 1]
        y_up <- y[ , 2]
        y <- NA
    } else {
        y_low <- NA
        y_up <- NA
    }

    drx <- unlist(Xlim[2, ]) - unlist(Xlim[1, ])

    ## define X & y labels
    if (is.null(ylab) && !is.null(names(y))) ylab <- names(y)[1]
    if (is.null(Xlab)) Xlab <- names(X)

    ## define X & y labels
    if (is.null(ylab)) ylab <- "y"
    if (is.null(Xlab)) Xlab <- paste(sep = "", "X", 1:D)

    fcenter <- tryFormat(x = center, drx = drx)

    ## Each 'id' will produce a plot
    for (id in 1:dim(axis)[1]) {
        if (D>1) screen(id, new=!add)

        d <- axis[id,]

        xlim = Xlim[, d]

        if (is.null(title)){
            title_d <- paste(collapse = "~",sep = "~", ylab, Xlab[d])
            if (D>1) {
                title_d <-  paste(collapse =title_sep, sep =title_sep, title_d, paste(Xlab[-d], '=', fcenter[-d]))
            }
        } else {
            title_d <- title
        }

        ## fading colors for points
        if (D>1) {
            xrel <- scale(x = as.matrix(X),
                          center = center,
                          scale = drx)

            ## ind.nonfix flags the non fixed dims
            ind.nonfix <- (1:D) %in% d[1] #c(d[1], d[2])
            ind.nonfix <- !ind.nonfix

            alpha <- pmax(0,apply(X = xrel[ , ind.nonfix, drop = FALSE],
                           MARGIN = 1,
                           FUN = function(x) (1 - sqrt(sum(x^2)/D))^bg_fading))
        } else {
            alpha <- rep(1, n)
        }

        if (isTRUE(add) && !all(is.na(y))) {
            if (D>1)
                plot(x=X[,d], y=y, # Cannot use 'points' so use 'plot' with these neutral args
                       col = fade(color = col_points, alpha = alpha),
                       pch = 20,type='p',
                       xlab="",ylab="", main="", xlim=xlim, ylim=ylim,
                     bty='n', xaxt='n', yaxt='n', ann=FALSE, # remove all text, that should be already displayed
                     ...)
            else
                points(x=X[,d], y=y,
                 col = fade(color = col_points, alpha = alpha),
                 pch = 20,type='p',
                 xlab="",ylab="", xlim=xlim, ylim=ylim,
                 ...)
        } else {
            plot(X[,d], if (!all(is.na(y))) y else y_low,
                 xlab=Xlab[d], ylab=ylab, xlim=xlim, ylim=ylim,
                 main=title_d,
                 pch = 20, type = if (!all(is.na(y))) "l" else "n",
                 col = fade(color = col_points, alpha = alpha),
                 ...)
            if(D>1) abline(v=center[d],col=col_center,lty=lty_center)
        }

        if (!all(is.na(y_low)) && !all(is.na(y_up))) {
            #for (p in 1:length(conf_level)) {
                for (i in 1:n) {
                        lines(x=c(X[i,d],X[i,d]),
                              y=c(y_low[i], y_up[i]),
                              col = fade(color = col_points, alpha = alpha[i]*col_fading_interval),
                              lwd = 5, lend = 1)
                }
            #}
        }
    }
}


#' @param eval_str the expression to evaluate in each subplot.
#' @param axis optional matrix of 2-axis combinations to plot, one by row. The value \code{NULL} leads to all possible combinations i.e. \code{choose(D, 2)}.
#' @param mfrow  an optional list to force \code{par(mfrow = ...)} call. The default value  \code{NULL} is automatically set for compact view.
#' @rdname sectionview
#' @method sectionview character
#' @aliases sectionview,character,character-method
#' @export
#' @seealso \code{\link{sectionview.matrix}} for a section plot, and \code{\link{sectionview3d.matrix}} for a 2D section plot.
#' @examples
#' x1 <- rnorm(15)
#' x2 <- rnorm(15)
#'
#' y <- x1 + x2^2 + rnorm(15)
#' model <- glm(y ~ x1 + I(x2^2))
#'
#' sectionview(model, center=c(.5,.5))
#'
#' sectionview("abline(h=5)")
sectionview.character <- function(eval_str,
                              axis = NULL,
                              mfrow = NULL,
                              ...) {

    if (!exists(".split.screen.lim",envir=DiceView.env))
        stop(paste0("Cannot eval '",eval_str,"' when no previous sectionview() was called."))

    # setup for center/Xlim/axis/mfrow
    view = getView(1, TRUE, NA, NA, NA, axis, mfrow)
    D = view$D
    center = view$center
    Xlim = view$Xlim
    ylim = view$ylim
    axis = view$axis
    mfrow = view$mfrow

    ## Each 'id' will produce a plot
    for (id in 1:dim(axis)[1]) {

        d <- axis[id,]

        e = parent.frame()
        assign("d",d,envir=e)
        assign("xlim",Xlim[,d],envir=e)
        assign("ylim",ylim,envir=e)

        if (D>1) {
            screen(id, new=FALSE)
            plot(x=Xlim[,d], y=ylim,
                 type='n',
                 xlab="",ylab="", main="",
                 bty='n', xaxt='n', yaxt='n', ann=FALSE, # remove all text, that should be already displayed
                 ...)
        }

        eval(parse(text = eval_str), envir = e)
    }
}

#' @param km_model an object of class \code{"km"}.
#' @param type the kriging type to use for model prediction.
#' @param col_points color of points.
#' @param conf_level confidence intervals to display.
#' @param conf_fading an optional factor of alpha (color channel) fading used to plot confidence intervals.
#' @param bg_fading  an optional factor of alpha (color channel) fading used to plot design points outside from this section.
#' @template sectionview-doc
#' @rdname sectionview
#' @method sectionview km
#' @aliases sectionview,km,km-method
#' @export
#' @seealso \code{\link{sectionview.km}} for a section plot, and \code{\link{sectionview3d.km}} for a 2D section plot.
#' @examples
#' if (requireNamespace("DiceKriging")) { library(DiceKriging)
#'
#' X = matrix(runif(15*2),ncol=2)
#' y = apply(X,1,branin)
#'
#' model <- km(design = X, response = y, covtype="matern3_2")
#'
#' sectionview(model, center=c(.5,.5))
#'
#' }
#'
sectionview.km <- function(km_model, type = "UK",
                           center = NULL,
                           axis = NULL,
                           npoints = 101,
                           col_points = if (!is.null(col)) col else "red",
                           col_fun = if (!is.null(col)) col else "blue",
                           col = NULL,
                           conf_level = 0.95,
                           conf_fading = 0.5,
                           bg_fading = 5,
                           mfrow = NULL,
                           Xlab = NULL, ylab = NULL,
                           Xlim = NULL, ylim=NULL,
                           title = NULL, title_sep = " | ",
                           add = FALSE,
                           ...) {
    X_doe <- km_model@X
    y_doe <- km_model@y

    D <- ncol(X_doe)
    n <- nrow(X_doe)

    if (km_model@noise.flag) {
        sdy_doe <- sqrt(km_model@noise.var)
    } else if (km_model@covariance@nugget.flag) {
        sdy_doe <- rep(sqrt(km_model@covariance@nugget), n)
    } else {
        sdy_doe <- 0
    }

    ## find limits
    if (is.null(Xlim) && !add)
        Xlim <- apply(X_doe, 2, range)
    else if (!is.null(Xlim))
        Xlim <- matrix(Xlim,nrow=2,ncol=D)

    if (is.null(ylim) && !add) {
        ymin <- min(y_doe-3*sdy_doe)
        ymax <- max(y_doe+3*sdy_doe)
        ylim <- c(ymin, ymax)
    }

    ## define X & y labels
    if (is.null(ylab)) ylab <- names(y_doe)
    if (is.null(Xlab)) Xlab <- names(X_doe)

    if (is.null(conf_fading) || length(conf_fading) != length(conf_level)) {
        conf_fading <- rep(0.5/length(conf_level), length(conf_level))
    }

    # plot mean
    sectionview.function(fun = function(x) {
        DiceKriging::predict.km(km_model,type=type,newdata=x,checkNames=FALSE)$mean
    }, vectorized=TRUE,
    center = center, axis = axis, mfrow = mfrow, Xlim = Xlim, ylim=ylim,
    npoints = npoints,
    col_fun = col_fun, #conf_fading=conf_fading,
    Xlab = Xlab, ylab = ylab,
    title = title, title_sep  = title_sep, add = add, ...)

    # plot design points
    sectionview.matrix(X = X_doe, y = y_doe,
                       col_points = col_points,
                       col_fading_interval = conf_fading, bg_fading = bg_fading,
                       add=TRUE)

    # plot confidence bands
    for (l in conf_level) {
        sectionview.function(fun = function(x) {
                p = DiceKriging::predict.km(km_model,type=type,newdata=x,checkNames=FALSE)
                cbind(p$mean-qnorm(1-(1-l)/2) * p$sd, p$mean+qnorm(1-(1-l)/2) * p$sd)
            }, vectorized=TRUE,
            npoints = npoints,
            col_fun = col_fun,
            col_fading_interval=conf_fading,
            add = TRUE)

        if (km_model@noise.flag)
            sectionview.matrix(X = X_doe, y = cbind(y_doe-qnorm(1-(1-l)/2) * sdy_doe, y_doe+qnorm(1-(1-l)/2) * sdy_doe),
                           col_points = col_points,
                           col_fading_interval = conf_fading, bg_fading = bg_fading,
                           add=TRUE)
    }

}

#' @param libKriging_model an object of class \code{"Kriging"} or \code{"WarpKriging"}.
#' @param col_points color of points.
#' @param bg_fading  an optional factor of alpha (color channel) fading used to plot design points outside from this section.
sectionview_libKriging <- function(libKriging_model,
                           center = NULL,
                           axis = NULL,
                           npoints = 101,
                           col_points = if (!is.null(col)) col else "red",
                           col_fun = if (!is.null(col)) col else "blue",
                           col = NULL,
                           conf_level = 0.95,
                           conf_fading = 0.5,
                           bg_fading = 5,
                           mfrow = NULL,
                           Xlab = NULL, ylab = NULL,
                           Xlim = NULL, ylim=NULL,
                           title = NULL, title_sep = " | ",
                           add = FALSE,
                           ...) {
    X_doe <- libKriging_model$X()
    y_doe <- libKriging_model$y()

    D <- ncol(X_doe)
    n <- nrow(X_doe)

    if (inherits(libKriging_model, "Kriging")) {
        nm <- libKriging_model$noise_model()
        if (nm == "nugget") {
            sdy_doe <- rep(sqrt(libKriging_model$nugget()), n)
        } else if (nm == "heterogeneous") {
            sdy_doe <- sqrt(libKriging_model$noise())
        } else {
            sdy_doe <- rep(0, n)
        }
    } else if (inherits(libKriging_model, "WarpKriging")) {
        sdy_doe <- rep(0, n)
    } else
        stop(paste0("Kriging model of class ",class(libKriging_model)," is not yet supported."))

    ## find limits
    if (is.null(Xlim) && !add)
        Xlim <- apply(X_doe, 2, range)
    else if (!is.null(Xlim))
        Xlim <- matrix(Xlim,nrow=2,ncol=D)

    if (is.null(ylim) && !add) {
        if (is.null(sdy_doe))
            ylim = range(y_doe)
        else {
            ymin <- min(y_doe-3*sdy_doe)
            ymax <- max(y_doe+3*sdy_doe)
            ylim <- c(ymin, ymax)
        }
    }

    ## define X & y labels
    if (is.null(ylab)) ylab <- names(y_doe)
    if (is.null(ylab)) ylab <- "y"
    if (is.null(Xlab)) Xlab <- names(X_doe)
    if (is.null(Xlab)) Xlab <- paste(sep = "", "X", 1:D)

    if (is.null(conf_fading) ||
        length(conf_fading) != length(conf_level))
        conf_fading <- rep(0.5/length(conf_level), length(conf_level))

    # plot mean
    sectionview.function(fun = function(x) {
        rlibkriging::predict(libKriging_model,x,return_stdev=FALSE)$mean
    }, vectorized=TRUE,
    center = center, axis = axis, mfrow = mfrow, Xlim = Xlim, ylim=ylim,
    npoints = npoints,
    col_fun = col_fun, #conf_fading=conf_fading,
    Xlab = Xlab, ylab = ylab,
    title = title, title_sep = title_sep, add = add, ...)

    # plot design points
    sectionview.matrix(X = X_doe, y = y_doe,
                       col_points = col_points,
                       col_fading_interval = conf_fading, bg_fading = bg_fading,
                       add=TRUE)

    # plot confidence bands
    for (l in conf_level) {
        sectionview.function(fun = function(x) {
                p = rlibkriging::predict(libKriging_model,x,return_stdev=TRUE)
                cbind(p$mean-qnorm(1-(1-l)/2) * p$stdev, p$mean+qnorm(1-(1-l)/2) * p$stdev)
            }, vectorized=TRUE,
            npoints = npoints,
            col_fun = col_fun,
            col_fading_interval=conf_fading,
            add = TRUE)

        if (any(sdy_doe != 0)) # so sdy_doe!=0
            sectionview.matrix(X = X_doe, y = cbind(y_doe-qnorm(1-(1-l)/2) * sdy_doe, y_doe+qnorm(1-(1-l)/2) * sdy_doe),
                           col_points = col_points,
                           col_fading_interval = conf_fading, bg_fading = bg_fading,
                           add=TRUE)
    }

}

#' @param Kriging_model an object of class \code{"Kriging"}.
#' @param col_points color of points.
#' @param conf_level confidence intervals to display.
#' @param conf_fading an optional factor of alpha (color channel) fading used to plot confidence intervals.
#' @param bg_fading  an optional factor of alpha (color channel) fading used to plot design points outside from this section.
#' @template sectionview-doc
#' @rdname sectionview
#' @method sectionview Kriging
#' @aliases sectionview,Kriging,Kriging-method
#' @export
#' @seealso \code{\link{sectionview.Kriging}} for a section plot, and \code{\link{sectionview3d.Kriging}} for a 2D section plot.
#' @examples
#' if (requireNamespace("rlibkriging")) { library(rlibkriging)
#'
#' X = matrix(runif(15*2),ncol=2)
#' y = apply(X,1,branin)
#'
#' model <- Kriging(X = X, y = y, kernel="matern3_2")
#'
#' sectionview(model, center=c(.5,.5))
#'
#' }
#'
sectionview.Kriging <- function(Kriging_model,
                                   center = NULL,
                                   axis = NULL,
                                   npoints = 101,
                                   col_points = if (!is.null(col)) col else "red",
                                   col_fun = if (!is.null(col)) col else "blue",
                                   col = NULL,
                                   conf_level = 0.95,
                                   conf_fading = 0.5,
                                   bg_fading = 5,
                                   mfrow = NULL,
                                   Xlab = NULL, ylab = NULL,
                                   Xlim = NULL, ylim=NULL,
                                   title = NULL, title_sep = " | ",
                                   add = FALSE,
                                   ...) {
    sectionview_libKriging(Kriging_model,center,axis,npoints,
                           col_points=col_points,col_fun=col_fun,col=col,
                           conf_level,conf_fading,bg_fading,
                           mfrow,Xlab, ylab,Xlim,ylim,title,title_sep,add,...)
}

#' @param WarpKriging_model an object of class \code{"WarpKriging"}.
#' @param col_points color of points.
#' @param conf_level an optional list of confidence intervals to display.
#' @param conf_fading an optional factor of alpha (color channel) fading used to plot confidence intervals.
#' @param bg_fading  an optional factor of alpha (color channel) fading used to plot design points outside from this section.
#' @template sectionview-doc
#' @rdname sectionview
#' @method sectionview WarpKriging
#' @aliases sectionview,WarpKriging,WarpKriging-method
#' @export
#' @seealso \code{\link{sectionview.Kriging}} for a section plot, and \code{\link{sectionview3d.WarpKriging}} for a 2D section plot.
#' @examples
#' if (requireNamespace("rlibkriging")) { library(rlibkriging)
#'
#' X = matrix(runif(15*2),ncol=2)
#' y = apply(X,1,branin) + 5*rnorm(15)
#'
#' model <- WarpKriging(y = y, X = X, warping = c("affine","affine"), kernel="matern3_2")
#'
#' sectionview(model, center=c(.5,.5))
#'
#' }
#'
sectionview.WarpKriging <- function(WarpKriging_model,
                                center = NULL,
                                axis = NULL,
                                npoints = 101,
                                col_points = if (!is.null(col)) col else "red",
                                col_fun = if (!is.null(col)) col else "blue",
                                col = NULL,
                                conf_level = 0.95,
                                conf_fading = 0.5,
                                bg_fading = 5,
                                mfrow = NULL,
                                Xlab = NULL, ylab = NULL,
                                Xlim = NULL, ylim=NULL,
                                title = NULL, title_sep = " | ",
                                add = FALSE,
                                ...) {
    sectionview_libKriging(WarpKriging_model,center,axis,npoints,
                           col_points=col_points,col_fun=col_fun,col=col,
                           conf_level,conf_fading,bg_fading,
                           mfrow,Xlab, ylab,Xlim,ylim,title,title_sep,add,...)
}

#' @param glm_model an object of class \code{"glm"}.
#' @param col_points color of points.
#' @param conf_level an optional list of confidence intervals to display.
#' @param conf_fading an optional factor of alpha (color channel) fading used to plot confidence intervals.
#' @param bg_fading  an optional factor of alpha (color channel) fading used to plot design points outside from this section.
#' @template sectionview-doc
#' @rdname sectionview
#' @method sectionview glm
#' @aliases sectionview,glm,glm-method
#' @export
#' @seealso \code{\link{sectionview.glm}} for a section plot, and \code{\link{sectionview3d.glm}} for a 2D section plot.
#' @examples
#' x1 <- rnorm(15)
#' x2 <- rnorm(15)
#'
#' y <- x1 + x2^2 + rnorm(15)
#' model <- glm(y ~ x1 + I(x2^2))
#'
#' sectionview(model, center=c(.5,.5))
#'
sectionview.glm <- function(glm_model,
                           center = NULL,
                           axis = NULL,
                           npoints = 101,
                           col_points = if (!is.null(col)) col else "red",
                           col_fun = if (!is.null(col)) col else "blue",
                           col = NULL,
                           conf_level = 0.95,
                           conf_fading = 0.5,
                           bg_fading = 5,
                           mfrow = NULL,
                           Xlab = NULL, ylab = NULL,
                           Xlim = NULL, ylim=NULL,
                           title = NULL, title_sep = " | ",
                           add = FALSE,
                           ...) {

    # Get X & y labels
    if (is.null(Xlab)) Xlab <- all.vars(glm_model$formula)[-1] # assume just one y
    if (is.null(ylab)) {
        factors_names = all.vars(glm_model$formula)
        for (n in Xlab) {
            factors_names = factors_names[-which(factors_names==n)]
        }
        ylab = factors_names
    }

    D <- length(Xlab)
    n <- length(glm_model$residuals)

    X_doe <- do.call(cbind,lapply(Xlab,function(Xn)glm_model$data[[Xn]]))
    colnames(X_doe) <- Xlab
    y_doe <- do.call(cbind,lapply(ylab,function(yn)glm_model$data[[yn]]))
    colnames(y_doe) <- ylab

    ## find limits
    if (is.null(Xlim) && !add)
        Xlim <- apply(X_doe, 2, range)
    else if (!is.null(Xlim))
        Xlim <- matrix(Xlim,nrow=2,ncol=D)

    if (is.null(conf_fading) ||
        length(conf_fading) != length(conf_level))
        conf_fading <- rep(0.5/length(conf_level), length(conf_level))

    # plot mean
    sectionview.function(fun = function(x) {
        x = as.data.frame(x)
            colnames(x) <- Xlab
            predict.glm(glm_model, newdata=x, se.fit=FALSE)
    }, vectorized=TRUE,
    center = center,axis = axis, mfrow = mfrow, Xlim = Xlim, ylim=range(y_doe),
    npoints = npoints,
    col_fun = col_fun,
    Xlab = Xlab, ylab = ylab,
    title = title, title_sep  = title_sep, add = add, ...)

    # plot design points
    sectionview.matrix(X = X_doe, y = y_doe,
                   col_points = col_points,
                   col_fading_interval = conf_fading, bg_fading = bg_fading,
                   add=TRUE)

    # plot confidence bands
    for (l in conf_level) {
        sectionview.function(fun = function(x) {
                x = as.data.frame(x)
                colnames(x) <- Xlab
                p = predict.glm(glm_model, newdata=x, se.fit=TRUE)
                cbind(p$fit-qnorm(1-(1-l)/2) * p$se.fit, p$fit+qnorm(1-(1-l)/2) * p$se.fit)
            }, vectorized=TRUE,
            npoints = npoints,
            col_fun = col_fun,
            col_fading_interval=conf_fading,
            add = TRUE)
    }
}

#' @param modelFit_model an object returned by DiceEval::modelFit.
#' @param col_points color of points.
#' @param bg_fading  an optional factor of alpha (color channel) fading used to plot design points outside from this section.
#' @template sectionview-doc
#' @rdname sectionview
#' @method sectionview list
#' @aliases sectionview,list,list-method
#' @export
#' @seealso \code{\link{sectionview.glm}} for a section plot, and \code{\link{sectionview3d.glm}} for a 2D section plot.
#' @examples
#' if (requireNamespace("DiceEval")) { library(DiceEval)
#'
#' X = matrix(runif(15*2),ncol=2)
#' y = apply(X,1,branin)
#'
#' model <- modelFit(X, y, type = "StepLinear")
#'
#' sectionview(model, center=c(.5,.5))
#'
#' }
#'
sectionview.list <- function(modelFit_model,
                            center = NULL,
                            axis = NULL,
                            npoints = 101,
                            col_points = if (!is.null(col)) col else "red",
                            col_fun = if (!is.null(col)) col else "blue",
                            col = NULL,
                            bg_fading = 5,
                            mfrow = NULL,
                            Xlab = NULL, ylab = NULL,
                            Xlim = NULL, ylim=NULL,
                            title = NULL, title_sep = " | ",
                            add = FALSE,
                            ...) {

    X_doe <- modelFit_model$data$X
    y_doe <- modelFit_model$data$Y

    D <- ncol(X_doe)
    n <- nrow(X_doe)

    ## define X & y labels
    if (is.null(ylab)) ylab <- names(y_doe)
    if (is.null(ylab)) ylab <- "y"
    if (is.null(Xlab)) Xlab <- names(X_doe)
    if (is.null(Xlab)) Xlab <- paste(sep = "", "X", 1:D)

    ## find limits
    if (is.null(Xlim) && !add)
        Xlim <- apply(X_doe, 2, range)
    else if (!is.null(Xlim))
        Xlim <- matrix(Xlim,nrow=2,ncol=D)

    # plot mean
    sectionview.function(fun = function(x) {
        x = as.data.frame(x)
            colnames(x) <- Xlab
            DiceEval::modelPredict(modelFit_model, x)
    }, vectorized=TRUE,
    center = center,axis = axis,mfrow = mfrow,Xlim = Xlim, ylim=range(y_doe),
    npoints = npoints,
    col_fun = col_fun, #conf_fading=conf_fading,
    Xlab = Xlab, ylab = ylab,
    title = title, title_sep = title_sep, add = add, ...)

    # plot design points
    sectionview.matrix(X = X_doe, y = y_doe,
                   col_points = col_points,
                   bg_fading = bg_fading,
                   add=TRUE)
}



#### Wrapper for sectionview ####

#' @import methods
if(!isGeneric("sectionview")) {
    setGeneric(name = "sectionview",
               def = function(...) standardGeneric("sectionview")
    )
}

#' @title Plot a section view of a prediction model or function, including design points if available.
#' @details If available, experimental points are plotted with fading colors. Points that fall in the specified section (if any) have the color specified \code{col_points} while points far away from the center have shaded versions of the same color. The amount of fading is determined using the Euclidean distance between the plotted point and \code{center}.
#' @param ... arguments of the \code{sectionview.km}, \code{sectionview.glm}, \code{sectionview.Kriging} or \code{sectionview.function} function
#' @export
#' @examples
#' ## A 2D example - Branin-Hoo function
#' sectionview(branin, center= c(.5,.5), col='black')
#'
#' \dontrun{
#' ## a 16-points factorial design, and the corresponding response
#' d <- 2; n <- 16
#' design.fact <- expand.grid(seq(0, 1, length = 4), seq(0, 1, length = 4))
#' design.fact <- data.frame(design.fact); names(design.fact) <- c("x1", "x2")
#' y <- branin(design.fact); names(y) <- "y"
#'
#' if (requireNamespace("DiceKriging")) { library(DiceKriging)
#' ## model: km
#' model <- DiceKriging::km(design = design.fact, response = y)
#' sectionview(model, center= c(.5,.5))
#' sectionview(branin, center= c(.5,.5), col='red', add=TRUE)
#' }
#'
#' if (requireNamespace("rlibkriging")) { library(rlibkriging)
#' ## model: Kriging
#' model <- Kriging(X = as.matrix(design.fact), y = as.matrix(y), kernel="matern3_2")
#' sectionview(model, center= c(.5,.5))
#' sectionview(branin, center= c(.5,.5), col='red', add=TRUE)
#' }
#'
#' ## model: glm
#' model <- glm(y ~ 1+ x1 + x2 + I(x1^2) + I(x2^2) + x1*x2, data=cbind(y,design.fact))
#' sectionview(model, center= c(.5,.5))
#' sectionview(branin, center= c(.5,.5), col='red', add=TRUE)
#'
#' if (requireNamespace("DiceEval")) { library(DiceEval)
#' ## model: StepLinear
#' model <- modelFit(design.fact, y, type = "StepLinear")
#' sectionview(model, center= c(.5,.5))
#' sectionview(branin, center= c(.5,.5), col='red', add=TRUE)
#' }
#' }
#'
sectionview <- function(...){
    UseMethod("sectionview")
}

#' @importFrom utils installed.packages
.load3d = function(package = NULL) {
    if (!is.null(package)) { # prefered package defined
        if (!(package %in% rownames(installed.packages()))) {
            warning(paste0("Package ",package," should be available to use any DiceView::*3d functions."))
            return(NULL)
        } else {
            env3d$package <- package
            return(package)
        }
    } else { # auto select 3d package
        if (!is.null(env3d$package) && (env3d$package %in% rownames(installed.packages()))) {
            return(env3d$package)
        } else {
            packages = c("rgl","scatterplot3d")
            for (package in packages) {
                if (package %in% rownames(installed.packages())) {
                    return(package)
                }
            }
            warning(paste0("Cannot find any installed package for 3D: ",paste0(collapse=" or ",packages)))
            return(NULL)
        }
    }
}

env3d <- new.env()

.open3d = function(...) {
    package = .load3d()
    if (is.null(package)) {
        stop("No 3D package available.")
    } else if (package=="rgl") {
        rgl::open3d(...)
    } else if (package=="scatterplot3d") {
        return()
    } else stop(paste0("Unsupported 3D package: ",package))
}

#' @examples .plot3d(runif(10),runif(10),runif(10))
.plot3d = function(x, y=NULL, z=NULL,  col='black', alpha=0.5, add = FALSE, package = .load3d(),...) {
    if (is.null(y) & is.null(z)) {y = x[,2]; z = x[,3]; x=x[,1]}
    if (is.null(package)) {
        stop("No 3D package available.")
    } else if (package=="rgl") {
        rgl::plot3d(x=x, y=y, z=z, col=col,...)
    } else if (package=="scatterplot3d") {
        if (!add) {
            p3d = scatterplot3d::scatterplot3d(x=x, y=y, z=z, color=translude(col,alpha),...)
            assign(".p3d",value=p3d,envir=env3d)
        } else {
            p3d = get(".p3d",envir = env3d)
            p3d$points3d(x=x, y=y, z=z, color=translude(col,alpha),...)
        }
        return(p3d)
    } else stop(paste0("Unsupported 3D package: ",package))
}

#' @examples .plot3d(runif(10),runif(10),runif(10)); .points3d(runif(10),runif(10),runif(10),col='red')
.points3d = function(x, y=NULL, z=NULL,  col='black', alpha=0.5, box=FALSE, package = .load3d(), ...) {
    if (is.null(y) & is.null(z)) {y = x[,2]; z = x[,3]; x=x[,1]}
    if (is.null(package)) {
        stop("No 3D package available.")
    } else if (package=="rgl") {
        rgl::points3d(x, y , z ,  color=col, alpha,...)
    } else if (package=="scatterplot3d") {
        p3d = get(".p3d",envir = env3d)
        p3d$points3d(x=x, y=y, z=z, col=translude(col,alpha),...)
    } else stop(paste0("Unsupported 3D package: ",package))
}

#' @examples .plot3d(runif(10),runif(10),runif(10)); .lines3d(runif(10),runif(10),runif(10),col='red')
.lines3d = function(x, y=NULL, z=NULL,  col='black' , alpha=0.5, box=FALSE, package = .load3d(),...) {
    if (is.null(package)) {
        stop("No 3D package available.")
    } else if (package=="rgl") {
        rgl::lines3d(x, y , z ,  color=col , alpha,...)
    } else if (package=="scatterplot3d") {
        p3d = get(".p3d",envir = env3d)
        p3d$points3d(x=x, y=y, z=z, col=translude(col,alpha),type='l',...)
    } else stop(paste0("Unsupported 3D package: ",package))
}

#' @examples .plot3d(runif(10),runif(10),runif(10)); .triangles3d(runif(3),runif(3),runif(3),col='red')
.triangles3d = function(x, y=NULL, z=NULL, col='black', alpha=0.5, box=FALSE, package = .load3d(), ...) {
    if (is.null(y) & is.null(z)) {
        if (is.matrix(x)) {
            y = x[,2]; z = x[,3]; x=x[,1]
        } else {
            y = x[2]; z = x[3]; x=x[1]
        }
    }
    if (is.null(package)) {
        stop("No 3D package available.")
    } else if (package=="rgl") {
        rgl::triangles3d(x ,y , z ,  color=col , alpha,... )
    } else if (package=="scatterplot3d") {
        p3d = get(".p3d",envir = env3d)
        polygon(p3d$xyz.convert(x,y,z),col = translude(col,alpha), border=translude(col,alpha),...)
    } else stop(paste0("Unsupported 3D package: ",package))
}

#' @examples .plot3d(runif(10),runif(10),runif(10)); x=runif(4); y=runif(4); z=runif(4); .quads3d(x,y,z,col='red'); p=c(4,3,2,1); .quads3d(x[p],y[p],z[p],col='red')
.quads3d = function(x, y=NULL, z=NULL, col='black', alpha=0.5, box=FALSE, package = .load3d(), ...) {
    if (is.null(y) & is.null(z)) {
        if (is.matrix(x)) {
            y = x[,2]; z = x[,3]; x=x[,1]
        } else {
            y = x[2]; z = x[3]; x=x[1]
        }
    }
    if (is.null(package)) {
        stop("No 3D package available.")
    } else if (package=="rgl") {
        rgl::quads3d(x ,y , z ,  color=col , alpha,... )
    } else if (package=="scatterplot3d") {
        p3d = get(".p3d",envir = env3d)
        polygon(p3d$xyz.convert(x,y,z),col = translude(col,alpha), border=translude(col,alpha),...)
    } else stop(paste0("Unsupported 3D package: ",package))
}

.surface3d = function(x, y, z, col='black', alpha = 0.5, box=FALSE, package = .load3d(), ...) {
    if (is.null(package)) {
        stop("No 3D package available.")
    } else if (package=="rgl") {
        rgl::surface3d(x ,y , z ,  color=col , alpha,... )
    } else if (package=="scatterplot3d") {
        p3d = get(".p3d",envir = env3d)
        # if (nrow(z)!=length(y)) z = t(z)
        # if (ncol(z)!=length(x)) z = matrix(z,nrow=length(y),ncol=length(x))
        # stop(paste0("z should be ",length(y),"x",length(x)," or ",length(x),"x",length(y)))
        for (i in 1:length(x)) {
            p3d$points3d(x=rep(x[i],length(y)), y=y, z=z[ (1:length(y)) * length(x) + (i-length(x))], col=translude(col,alpha),type='l',...)
        }
        for (j in 1:length(y)) {
            p3d$points3d(x=x,y=rep(y[j],length(x)), z=z[ ((j-1)*length(x)+1):(j*length(x)) ], col=translude(col,alpha),type='l',...)
        }
    } else stop(paste0("Unsupported 3D package: ",package))
}

##### Tests #####
# X = matrix(runif(40),ncol=2)
# p = 2
# f <- function(X) apply(X, 1,
#                        function(x)
#                            prod(
#                                sin(2*pi*
#                                        ( x * (seq(0,1,l=1+length(x))[-1])^p )
#                                )))
# Y = f(X)
#
# .plot3d(x=X[,1],y=X[,2],z=Y)
#
# X2 = matrix(runif(40),ncol=2)
# Y2 = f(X2)
# .points3d(x=X2[,1],y=X2[,2],z=Y2,col='red')
# .lines3d(x=X2[,1],y=X2[,2],z=Y2,col='red')
#
# X3 = matrix(runif(6),ncol=2)
# Y3 = f(X3)
# .triangles3d(x=X3[,1],y=X3[,2],z=Y3,col='red')
#
# X4_1 = seq(0,1,,21)
# X4_2 = seq(0,1,,31)
# Y4 = f(expand.grid(X4_1,X4_2))
# .surface3d(x=X4_1,y=X4_2,z=Y4,col='red')

##=========================================================
## Manage the DiceView environment for screen setup
##=========================================================

DiceView.env <- new.env()
# Preliminary setup for center/Xlim/axis/mfrow regarding add=TRUE/FALSE
getView <- function(viewdim, add, center, Xlim, ylim, axis, mfrow, verbose=FALSE) {
    if (!isTRUE(add)) {
        if (is.null(Xlim)) stop("Xlim must be specified.")

        if (is.null(center) && is.matrix(Xlim)) {
            center <- colMeans(Xlim, na.rm = TRUE)
        }

        if (is.null(center)) {
            warning("center not specified, assume 1D function")
            D <- 1
        } else
            D <- length(center)

        Xlim <- matrix(unlist(Xlim),nrow=2,ncol=D) # ensure not list + dup c(0,1) if needed
        if (!isTRUE(ncol(Xlim) == D))
            stop(paste0("Xlim must be a matrix with", D," columns."))
        if (!is.null(center)) # ensure center is included in Xlim
            for (i in 1:D) {
                Xlim[1,i] <- min(Xlim[1,i],center[i])
                Xlim[2,i] <- max(Xlim[2,i],center[i])
            }

        if (is.null(ylim)) ylim=c(NA,NA) # to be updated later

        close.screen( all.screens = TRUE )

        assign(".split.screen.lim",cbind(Xlim,ylim),envir=DiceView.env) # matrix of limits, each col for one dim
        assign(".split.screen.center",center,envir=DiceView.env)

        if (is.null(axis)) {
            axis <- t(utils::combn(D, viewdim))
        } else {
            axis <- matrix(axis, ncol = viewdim)
        }
        if (is.null(mfrow)) {
            nc <- round(sqrt(nrow(axis)))
            nl <- ceiling(nrow(axis)/nc)
            mfrow <- c(nc, nl)
        }

        if (D > viewdim) {
            split.screen(figs = mfrow, erase=TRUE)
        }

        assign(".split.screen.axis",axis,envir=DiceView.env)
        assign(".split.screen.mfrow",mfrow,envir=DiceView.env)
    } else {
        if (!exists(".split.screen.lim",envir=DiceView.env) || !exists(".split.screen.center",envir=DiceView.env))
            stop("Cannot use add=TRUE when no previous sectionview() was called.")

        # re-use global settings for limits of this screen
        Xlim.ignored = !is.null(Xlim)
        Xlim = get(x=".split.screen.lim",envir=DiceView.env)
        D <- ncol(Xlim)-1
        ylim = Xlim[,(D+1)] # get ylim
        Xlim = Xlim[,-(D+1),drop=FALSE] # remove ylim
        if (Xlim.ignored) warning(paste0("Xlim ignored when add=TRUE, replaced by: ",paste0(Xlim, collapse = ",")))

        center = get(x=".split.screen.center",envir=DiceView.env)
        axis = get(x=".split.screen.axis",envir=DiceView.env)
        mfrow = get(x=".split.screen.mfrow",envir=DiceView.env)
    }

    if (verbose) {
        cat("D: ", D, "\n")
        cat("center: ", paste0(collapse=",",center), "\n")
        cat("Xlim: ", paste0(collapse=",",Xlim), "\n")
        cat("ylim: ", paste0(collapse=",",ylim), "\n")
        cat("axis: ", paste0(collapse=",",axis), "\n")
        cat("mfrow: ", paste0(collapse=",",mfrow), "\n")
    }

    return(list(D=D, center=center, Xlim=Xlim, ylim=ylim, axis=axis, mfrow=mfrow))
}

# Force setup of some DiceView environment
setView_ylim = function(ylim) {
    if (exists(".split.screen.lim",envir=DiceView.env)) {
        Xlim = get(x=".split.screen.lim",envir=DiceView.env)
        D <- ncol(Xlim)-1
        Xlim = Xlim[,-(D+1)] # remove ylim
        assign(".split.screen.lim",cbind(Xlim,ylim),envir=DiceView.env)

        # cat("cbind(Xlim,ylim): ", paste0(collapse=",",get(x=".split.screen.lim",envir=DiceView.env), "\n"))
    } else {
        stop("Cannot use setView() when no previous DieceView env setup.")
    }
}


##=========================================================
## make colors semi-transparent
##
## alpha = 0 OPAQUE
## alpha = 1 FULLY TRANSPARENT
##
##=========================================================

#' @import grDevices
translude <- function(col, alpha = 0.6) {

    alpha <- rep(alpha, length.out = length(col))
    rgb <- as.matrix(col2rgb(col)/255)
    colors2 <- rgb(red = rgb["red", ],
                   green = rgb["green", ],
                   blue = rgb["blue", ],
                   alpha = alpha)


}

# check integer
is.wholenumber <- function(x, tol = .Machine$double.eps^0.5)  abs(x - round(x)) < tol

##========================================================
## level color: one base color incremented in hsv to provide a palette
##
##========================================================

#' @import grDevices
col.levels <- function(color, levels, fill=FALSE){
    if (length(levels)!=1) # if nlevels is in fact levels
        nlevels <- length(levels)
    else nlevels <- levels
    if (fill) nlevels <- nlevels - 1

    col.rgb=col2rgb(color, alpha = TRUE)
    col.hsv=rgb2hsv(r=col.rgb[1],g=col.rgb[2],b=col.rgb[3])
    col = hsv(h=col.hsv[1],s=seq(f=0,t=col.hsv[2],l=nlevels),v=col.hsv[3],alpha=col.rgb[4]/255)
    return(col)
}

##========================================================
## level colors: two base colors incermented in hsv to provide a palette
##
##========================================================

#' @import grDevices
cols.levels <- function(color1,color2,levels, fill=FALSE) {
    if (length(levels)!=1) # if nlevels is in fact levels
        nlevels <- length(levels)
    else nlevels <- levels
    if (fill) nlevels <- nlevels - 1

    col1.rgb=col2rgb(color1, alpha = TRUE)
    col2.rgb=col2rgb(color2, alpha = TRUE)
    col1.hsv=rgb2hsv(r=col1.rgb[1],g=col1.rgb[2],b=col1.rgb[3])
    col2.hsv=rgb2hsv(r=col2.rgb[1],g=col2.rgb[2],b=col2.rgb[3])
    col = hsv(h=seq(f=col1.hsv[1],t=col2.hsv[1],l=nlevels),
              s=seq(f=col1.hsv[2],t=col2.hsv[2],l=nlevels),
              v=seq(f=col1.hsv[3],t=col2.hsv[3],l=nlevels),
              alpha=seq(f=col1.rgb[4],t=col2.rgb[4],l=nlevels)/255)
    return(col)
}

##========================================================
## fade color: one color and several fading values
## between 0 and 1
##
## alpha ~ 0 nearly equal to the input 'color'
## alpha ~ 1 nearly invisible grayed/transparent
##
## The returned value is vector. It is build
## from a matrix with four rows as
## those produced by col2rgb(x, alpha = TRUE)
##
##========================================================

# be careful that fade(.., alpha=0) means total fading, while fade(.., alpha=1) means no fading
#' @import grDevices
fade <- function(color = "red",
                 alpha =  seq(from = 0, to = 1, length.out = 5),
                 plot = FALSE) {

    if (any(alpha < 0) || any(alpha > 1)) stop("'alpha' values must be >=0 and <= 1")
    if (length(color) > 1) stop("'color' must be of length 1")

    ## a matrix with  1 col
    rgbcol <- col2rgb(color)/255
    mat <- matrix(1-alpha, nrow = 3, ncol = length(alpha), byrow = TRUE)

    mat <- mat + rgbcol %*% alpha

    colors2 <- rgb(red = mat[1, ],
                   green = mat[2, ],
                   blue = mat[3, ],
                   alpha = alpha)

    if (plot) {
        x <- seq(from = 0, to = 1, length.out = length(alpha))
        plot.new( )
        for ( i in 1:(length(alpha)) ){
            rect(xleft = x[i],
                 xright = x[i+1],
                 ybottom = 0,
                 ytop =  1,
                 border = NA,
                 col = colors2[i])
        }

    }

    colors2

}

fades <- function(colors = c('red','blue'),
                  alpha =  0.5)
    Vectorize(function(col)fade(col,alpha = alpha,plot=FALSE))(colors)

##========================================================
## try to find a good formatted value for a numeric vector
## x using a vector of diff range drx
##
## For instance, if drx is 1000, no decimal or very few
## decimals should be used.
##
##========================================================

tryFormat <- function(x, drx) {
    d <- length(x)
    ldx <- log(drx, base = 10)
    ff <- rep(1, d)
    fd <- rep(1, d)
    ff[ldx > 0] <- ceiling(ldx[ldx > 0])
    fd[ldx < 0] <- ceiling(-ldx[ldx < 0]) + 2
    ff <- ff + fd +1

        formats <- paste("%", ff, ".", fd, "f", sep = "")
    fx = format(x)
    try(fx <- sprintf(formats, x))
    fx
}

#' This is a simple copy of the Branin-Hoo 2-dimensional test function, as provided in DiceKriging package.
#' The Branin-Hoo function is defined here over [0,1] x [0,1], instead of [-5,0] x [10,15] as usual.
#' It has 3 global minima : x1 = c(0.9616520, 0.15); x2 = c(0.1238946, 0.8166644); x3 = c(0.5427730, 0.15)
#'
#' @param x a 2-dimensional vector specifying the location where the function is to be evaluated.
#'
#' @return A real number equal to the Branin-Hoo function values at x
#' @export
branin <- function(x) {
    x1 <- x[1] * 15 - 5
    x2 <- x[2] * 15
    (x2 - 5/(4 * pi^2) * (x1^2) + 5/pi * x1 - 6)^2 + 10 * (1 - 1/(8 * pi)) * cos(x1) + 10
}

maxWorkers <- function() {
    chk <- Sys.getenv("_R_CHECK_LIMIT_CORES_", "")
    if (nzchar(chk) && chk == "TRUE") {
        return(2)
    } else {
        if (Sys.info()[["sysname"]] == "Windows") {
            return(1) #parallel::detectCores() - 1)
        } else {
            return(parallel::detectCores())
        }
    }
}

#### Generalization of root finding ####

# Brent's method root-finding (translated from C version in stats::R_zeroin2)
R_zeroin2 <- function(f, ax, bx, fa, fb, Tol, Maxit) {
  EPSILON <- .Machine$double.eps
  a <- ax
  b <- bx
  c <- a
  fc <- fa
  maxit <- Maxit + 1
  tol <- Tol

  if (fa == 0.0) {
    Tol <- 0.0
    Maxit <- 0
    return(a)
  }
  if (fb == 0.0) {
    Tol <- 0.0
    Maxit <- 0
    return(b)
  }

  while (maxit > 0) {
    maxit <- maxit - 1
    prev_step <- b - a
    tol_act <- 2 * EPSILON * abs(b) + tol / 2
    new_step <- (c - b) / 2

    if (abs(new_step) <= tol_act || fb == 0.0) {
      Maxit <- Maxit - maxit
      Tol <- abs(c - b)
      return(b)
    }

    if (abs(prev_step) >= tol_act && abs(fa) > abs(fb)) {
      cb <- c - b
      if (a == c) {
        # Linear interpolation
        t1 <- fb / fa
        p <- cb * t1
        q <- 1.0 - t1
      } else {
        # Quadratic inverse interpolation
        qv <- fa / fc
        t1 <- fb / fc
        t2 <- fb / fa
        p <- t2 * (cb * qv * (qv - t1) - (b - a) * (t1 - 1.0))
        q <- (qv - 1.0) * (t1 - 1.0) * (t2 - 1.0)
      }
      if (exists("q")) {
        if (p > 0) {
          q <- -q
        } else {
          p <- -p
        }
        if (p < (0.75 * cb * q - abs(tol_act * q) / 2) &&
            p < abs(prev_step * q / 2)) {
          new_step <- p / q
        }
      }
    }

    if (abs(new_step) < tol_act) {
      if (new_step > 0) {
        new_step <- tol_act
      } else {
        new_step <- -tol_act
      }
    }

    a <- b
    fa <- fb
    b <- b + new_step
    fb <- f(b)
    if ((fb > 0 && fc > 0) || (fb < 0 && fc < 0)) {
      c <- a
      fc <- fa
    }
  }

  Tol <- abs(c - b)
  Maxit <- -1
  return(b)
}

#' @title One Dimensional Root (Zero) Finding
#' @description Search one root with given precision (on y). Iterate over uniroot as long as necessary.
#' @param f the function for which the root is sought.
#' @param lower the lower end point of the interval to be searched.
#' @param upper the upper end point of the interval to be searched.
#' @param maxerror_f the maximum error on f evaluation (iterates over uniroot to converge).
#' @param f_lower the same as f(lower).
#' @param f_upper the same as f(upper).
#' @param tol the desired accuracy (convergence tolerance on f arg).
#' @param convexity the learned convexity factor of the function, used to reduce the boundaries for uniroot.
#' @param rec counter of recursive level.
#' @param max.rec maximal number of recursive level before failure (stop).
#' @param ... additional named or unnamed arguments to be passed to f.
#' @export
#' @author Yann Richet, ASNR
#' @examples
#' f=function(x) {cat("f");1-exp(x)}; f(root(f,lower=-1,upper=2))
#' f=function(x) {cat("f");exp(x)-1}; f(root(f,lower=-1,upper=2))
#'
#' .f = function(x) 1-exp(1*x)
#' f=function(x) {cat("f");y=.f(x);points(x,y,pch=20,col=rgb(0,0,0,.2));y}
#' plot(.f,xlim=c(-1,2)); f(root(f,lower=-1,upper=2))
#'
#' .f = function(x) exp(10*x)-1
#' f=function(x) {cat("f");y=.f(x);points(x,y,pch=20);y}
#' plot(.f,xlim=c(-1,2)); f(root(f,lower=-1,upper=2))
#'
#' .f = function(x) exp(100*x)-1
#' f=function(x) {cat("f");y=.f(x);points(x,y,pch=20);y}
#' plot(.f,xlim=c(-1,2)); f(root(f,lower=-1,upper=2))
#'
#' f=function(x) {cat("f");exp(100*x)-1}; f(root(f,lower=-1,upper=2))
#'
#' \dontrun{
#'
#'   # Quite hard functions to find roots
#'
#'   ## Increasing function
#'   ## convex
#'   n.f=0
#'   .f = function(x) exp(10*x)-1
#'   f=function(x) {n.f<<-n.f+1;y=.f(x);points(x,y,pch=20);y}
#'   plot(.f,xlim=c(-.1,.2)); f(root(f,lower=-1,upper=2))
#'   print(n.f)
#'   ## non-convex
#'   n.f=0
#'   .f = function(x) 1-exp(-10*x)
#'   f=function(x) {n.f<<-n.f+1;y=.f(x);points(x,y,pch=20);y}
#'   plot(.f,xlim=c(-.1,.2)); f(root(f,lower=-1,upper=2))
#'   print(n.f)
#'
#'   # ## Decreasing function
#'   # ## non-convex
#'   n.f=0
#'   .f = function(x) 1-exp(10*x)
#'   f=function(x) {n.f<<-n.f+1;y=.f(x);points(x,y,pch=20,col=rgb(0,0,0,.2));y}
#'   plot(.f,xlim=c(-.1,.2)); f(root(f,lower=-1,upper=2))
#'   print(n.f)
#'   # ## convex
#'   n.f=0
#'   .f = function(x) exp(-10*x)-1
#'   f=function(x) {n.f<<-n.f+1;y=.f(x);points(x,y,pch=20,col=rgb(0,0,0,.2));y}
#'   plot(.f,xlim=c(-.1,.2)); f(root(f,lower=-1,upper=2))
#'   print(n.f)
#' }
root <- function(f, lower, upper, maxerror_f = 1e-07,
                 f_lower = f(lower,  ...), f_upper = f(upper, ...),
                 tol = .Machine$double.eps^0.25,
                 convexity = FALSE, rec=0, max.rec=NA, ...) {
    # if (rec>50) warning(paste0("Many recursions:\n"," * lower=",lower,"\n"," * upper=",upper,"\n"," * f_lower=",f_lower,"\n"," * f_upper=",f_upper,"\n"," * convexity=",convexity,"\n"," * maxerror_f=",maxerror_f,"\n"," * tol=",tol,"\n"," * rec=",rec,"\n"))
    if (isTRUE(rec>max.rec)) {
      # x=seq(lower,upper,,11)
      # y=Vectorize(f)(x)
      # cat(paste0(apply(cbind(x,y),1,function(x) paste0(x,collapse=" -> ")),collapse="\n"))
      stop(paste0("Too many recursions:\n"," * lower=",lower,"\n"," * upper=",upper,"\n"," * f_lower=",f_lower,"\n"," * f_upper=",f_upper,"\n"," * convexity=",convexity,"\n"," * maxerror_f=",maxerror_f,"\n"," * tol=",tol,"\n"," * rec=",rec,"\n"))
    }

    tol = max(tol,.Machine$double.eps^0.25) # ensure suitable tol for later uniroot
    if (upper < lower) {
      lower.old = lower
      upper.old = upper
      f_lower.old = f_lower
      f_upper.old = f_upper
    return(root(f = f, lower = upper.old, upper = lower.old, maxerror_f = maxerror_f,
                #f_lower = f_upper.old, f_upper = f_lower.old,
                tol = tol, convexity = convexity, rec=rec+1, ...))
    }
    r = NULL
    ## Safe but slow
    #   try(r <- uniroot(f = f, lower = lower, upper = upper,
    #                  f.lower = f_lower, f.upper = f_upper,
    #                  tol = tol, ...)$root, silent = FALSE)
    ## Cannot use C function form stats, as raise WARNING for CRAN
    # r <- .Call(stats:::zeroin2, function(x) f(x, ...),
    #                 lower, upper, f.lower = f_lower, f.upper = f_upper, tol, 1000)[1]
    ## Use a R port instead
    r <- R_zeroin2(f = function(x) f(x, ...),
                        ax = lower, bx = upper,
                        fa = f_lower, fb = f_upper,
                        Tol = tol, Maxit = 1000)
    if (is.null(r)) {
        warning(paste0("No root found in [", lower, ",", upper, "] -> [", f_lower, ",", f_upper, "]"))
        return(NULL)
    }

## benchmark uniroot / C_zeroin2 : 6 times faster (!)
# f = function(x) x^2-1.2
# f_lower = f(0)
# f_upper = f(2)
# tol= 1e-5
# rbenchmark::benchmark(
#   brentDekker = pracma::brentDekker(f, a=0, b=2,
#                         # f.lower = f_lower, f.upper = f_upper,
#                         tol = tol, maxiter=1000),
#   brent = pracma::brent(f, a=0, b=2,
#                         # f.lower = f_lower, f.upper = f_upper,
#                         tol = tol, maxiter=1000),
#   fzero = pracma::fzero(f, x=1,
#                         # lower = 0, upper = 2,
#            # f.lower = f_lower, f.upper = f_upper,
#            tol = tol, maxiter=1000),
#   uniroot = uniroot(f = f, lower = 0, upper = 2,
#                     f.lower = f_lower, f.upper = f_upper,
#                     tol = tol, maxiter=1000,extendInt="no"),
#   C_zeroin2 = .External2(stats:::C_zeroin2, f,
#                           0, 2, f.lower = f_lower, f.upper = f_upper, tol, 1000)[1],
#   replications = 1000,
#   columns = c("test", "replications", "elapsed", "relative"),
#   order = "relative"
# )

    r_root = r
    f_root = f(r_root, ...)
    err = abs(f_root)/maxerror_f

    # print(paste0("[",lower,",",r_root,",",upper,"] -> ","[",f_lower,",",f_root,",",f_upper,"]"))

    if (!is.numeric(err)) stop(paste0("Error in root at ",r_root,": not numeric abs(", f_root,")/", maxerror_f, " = ", err))
    if (err > 1) {
      if (f_lower * f_root < 0) {

        x0 = r_root - (r_root - lower) * min(1-tol,max(tol,(f_root/(f_root - f_lower)) ^exp(convexity)))
        f_x0 = f(x0, ...)

        if (isTRUE(f_x0 * f_root < 0)) {
          return(root(f, lower = x0, upper = r_root, maxerror_f = maxerror_f,
                      f_lower = f_x0, f_upper = f_root,
                      convexity = if (!is.numeric(convexity)) convexity else convexity - log((f_x0 - f_lower)/(f_root - f_lower)),
                      tol = tol, rec=rec+1, max.rec=max.rec, ...))
        } else if (isFALSE(f_x0 * f_root < 0)) {
          return(root(f, lower = lower, upper = x0, maxerror_f = maxerror_f,
                      f_lower = f_lower, f_upper = f_x0,
                      convexity = if (!is.numeric(convexity)) convexity else convexity + log((f_x0 - f_lower)/(f_root - f_lower)),
                      tol = tol, rec=rec+1, max.rec=max.rec, ...))
        } else stop(paste0("Error in root at x0=",x0,", f(x0)=",f_x0," root=",r_root," f(root)=",f_root))

      } else if (f_upper * f_root < 0) {

        x0 = r_root + (upper - r_root) * min(1-tol,max(tol,(-f_root/(f_upper - f_root)) ^exp(convexity)))
        f_x0 = f(x0, ...)

        if (isTRUE(f_x0 * f_root < 0)) {
          return(root(f, lower = r_root, upper = x0, maxerror_f = maxerror_f,
                      f_lower = f_root, f_upper = f_x0,
                      convexity = if (!is.numeric(convexity)) convexity else convexity + log((f_upper - f_x0)/(f_upper - f_root)),
                      tol = tol, rec=rec+1, max.rec=max.rec, ...))
        } else if (isFALSE(f_x0 * f_root < 0)) {
          return(root(f, lower = x0, upper = upper, maxerror_f = maxerror_f,
                      f_lower = f_x0, f_upper = f_upper,
                      convexity = if (!is.numeric(convexity)) convexity else convexity - log((f_upper - f_x0)/(f_upper - f_root)),
                      tol = tol, rec=rec+1, max.rec=max.rec, ...))
        } else stop(paste0("Error in root at x0=",x0,", f(x0)=",f_x0," root=",r_root," f(root)=",f_root))

    } else stop(paste0("Error in root with lower=",lower," r_root=",r_root,", upper=",upper,", f_lower=",f_lower," f_root=",f_root,", f_upper=",f_upper))
  } else r_root
}

#' @title One Dimensional Multiple Roots (Zero) Finding
#' @description Search multiple roots of 1D function, sampled/splitted by a (1D) mesh
#' @param f Function to find roots
#' @param vectorized boolean: is f already vectorized ? (default: FALSE) or if function: vectorized version of f.
#' @param interval bounds to inverse in
#' @param split.size number of parts to perform uniroot inside
#' @param split function or "unif" or "seq" (default) to preform interval partition
#' @param maxerror_f the maximum error on f evaluation (iterates over uniroot to converge).
#' @param tol the desired accuracy (convergence tolerance on f arg).
#' @param .lapply control the loop/vectorization over different roots (defaults to multicore apply).
#' @param ... additional named or unnamed arguments to be passed to f.
#' @return array of x, so f(x)=target
#' @import stats
#' @export
#' @examples
#' roots(sin,interval=c(pi/2,5*pi/2))
#' roots(sin,interval=c(pi/2,1.5*pi/2))
#'
#' f=function(x)exp(x)-1;
#' f(roots(f,interval=c(-1,2)))
#'
#' f=function(x)exp(1000*x)-1;
#' f(roots(f,interval=c(-1,2)))
roots = function (f, vectorized=FALSE, interval, maxerror_f = 1e-07, split = "seq", split.size = 11,
          tol = .Machine$double.eps^0.25, .lapply = parallel::mclapply, ...) {
    lower = min(interval)
    upper = max(interval)

    if (is.function(split))
        intervals = split(min = lower, max = upper, n = split.size)
    else if (is.numeric(split))
        intervals = split
    else if (split == "seq")
        intervals = seq(from = lower, to = upper, length.out = split.size)
    else if (split == "unif")
        intervals = runif(min = lower, max = upper, n = split.size)
    else stop("unsupported mesh: function(min,max,n)/array/'seq'/'fun': type is ",
              typeof(split), " : ", paste0(split, collapse = ";", sep = ","))

    intervals <- sort(unique(intervals))

    if (isTRUE(vectorized))
        f_vec = function(x, ...) f(x, ...)
    else if (is.function(vectorized))
        f_vec = function(x, ...) vectorized(x, ...)
    else f_vec = Vectorize.function(f, dim = 1)

    f_intervals = f_vec(intervals)

    I.list <- lapply(seq_len(length(intervals) - 1),
                     function(i) if (f_intervals[i] * f_intervals[i + 1] < 0)
                                 c(intervals[i], intervals[i + 1], f_intervals[i] , f_intervals[i + 1])
                                 else NULL)

    I.list <- Filter(Negate(is.null), I.list)
    I.roots = .lapply(I.list, function(I) {
        r = NULL
        try({
            r = root(f, lower = I[1], upper = I[2], f_lower = I[3], f_upper = I[4],
                     maxerror_f = maxerror_f, tol = tol, ...)
        }, silent = F)
        r
    })
    I.roots = unlist(Filter(Negate(is.null), I.roots))

    if (length(I.roots) > 0)
        r = matrix(unlist(I.roots), ncol = 1)
    else
        r = NA

    attr(r, "mesh") <- list(p = matrix(intervals, ncol = 1))
    return(r)
}

#' @title Minimal distance between one point to many points
#' @param x one point
#' @param X matrix of points (same number of columns than x)
#' @param norm normalization vecor of distance (same number of columns than x)
#' @return minimal distance
#' @export
#' @examples
#' min_dist(runif(3),matrix(runif(30),ncol=3))
min_dist <- function (x, X, norm=rep(1,ncol(X))){
    return(min(sqrt(rowSums(((X - matrix(x, nrow = nrow(X), ncol = ncol(X), byrow = TRUE))/norm)^2))))
}

#' @title Multi Dimensional Multiple Roots (Zero) Finding, sampled by a mesh
#' @param f Function (one or more dimensions) to find roots of
#' @param vectorized boolean: is f already vectorized ? (default: FALSE) or if function: vectorized version of f.
#' @param intervals bounds to inverse in, each column contains min and max of each dimension
#' @param mesh.type function or "unif" or "seq" (default) to preform interval partition
#' @param mesh.sizes number of parts for mesh (duplicate for each dimension if using "seq")
#' @param vectorized vectorized f: function, TRUE (use f directly), or wrap in Vectorize.function: FALSE (default args), "lapply", "mclapply", ...
#' @param maxerror_f the maximum error on f evaluation (iterates over uniroot to converge).
#' @param tol the desired accuracy (convergence tolerance on f arg).
#' @param num_workers number of parallel roots finding
#' @param ... Other args for f
#' @import stats
#' @importFrom DiceDesign lhsDesign
#' @importFrom parallel mclapply
#' @export
#' @return matrix of x, so f(x)=0
#' @examples
#' roots_mesh(function(x) x-.51, intervals=rbind(0,1),
#'   num_workers=1)
#' roots_mesh(function(x) sum(x)-.51, intervals=cbind(rbind(0,1),rbind(0,1)),
#'   num_workers=1)
#' roots_mesh(sin,intervals=c(pi/2,5*pi/2),
#'   num_workers=1)
#' roots_mesh(f = function(x) sin(pi*x[1])*sin(pi*x[2]),
#'            intervals = matrix(c(1/2,5/2,1/2,5/2),nrow=2),
#'            num_workers=1)
#'
#' r = roots_mesh(f = function(x) (0.25+x[1])^2+(0.5+x[2])^2 - .25,
#' intervals=matrix(c(-1,1,-1,1),nrow=2), mesh.size=5,
#'   num_workers=1)
#' plot(r,xlim=c(-1,1),ylim=c(-1,1))
#'
#' r = roots_mesh(function(x) (0.5+x[1])^2+(-0.5+x[2])^2+(0.+x[3])^2 - .5,
#'                mesh.sizes = 11,
#'                intervals=matrix(c(-1,1,-1,1,-1,1),nrow=2),
#'                num_workers=1)
#' scatterplot3d::scatterplot3d(r,xlim=c(-1,1),ylim=c(-1,1),zlim=c(-1,1))
#'
#' roots_mesh(function(x)exp(x)-1,intervals=c(-1,2),
#'            num_workers=1)
#' roots_mesh(function(x)exp(1000*x)-1,intervals=c(-1,2),
#'            num_workers=1)
roots_mesh = function (f, vectorized = FALSE, intervals, mesh.type = "seq", mesh.sizes = 11,
                       maxerror_f = 1e-07, tol = .Machine$double.eps^0.25, num_workers=maxWorkers(), ...) {

  # .t1 <<- Sys.time()

  # if (is.matrix(intervals)) {
  #   lowers = apply(intervals, 2, min)
  #   uppers = apply(intervals, 2, max)
  #   d = ncol(intervals)
  # } else if (is.null(intervals) && is.matrix(mesh)) {
  #   lowers = apply(mesh, 2, min)
  #   uppers = apply(mesh, 2, max)
  #   d = ncol(mesh)
  # } else
  #   d = 1

  if (is.matrix(intervals))
    d = ncol(intervals)
  else if (is.null(intervals) && is.matrix(mesh.type))
    d = ncol(mesh)
  else
    d = 1

  if (d == 1) {
    r = roots(f, vectorized = vectorized, interval = intervals, split = mesh.type, split.size = mesh.sizes,
              maxerror_f = maxerror_f, tol = tol, ...)
    # warning(paste0("    [roots_mesh] roots d=1: ",Sys.time() - .t1))
    # .t1 <<- Sys.time()
    return(r)
  }

  # if (length(mesh.sizes) != d)
  #   mesh.size = rep(mesh.sizes, d)
  # if (is.matrix(mesh)) {
  #   ridge.points = mesh
  # } else if (is.function(mesh)) {
  #   ridge.points = mesh(prod(mesh.size), lowers, uppers)
  # } else if (mesh == "seq") {
  #   ridge.points = matrix(seq(f = lowers[1], to = uppers[1],
  #                             length.out = mesh.size[1]), ncol = 1)
  #   for (i in 2:d) {
  #     ridge.points = combn.design(ridge.points, matrix(seq(f = lowers[i], to = uppers[i], length.out = mesh.size[i]), ncol = 1))
  #   }
  # } else if (mesh == "unif") {
  #   ridge.points = matrix(runif(n = d * prod(mesh.size),
  #                               min = 0, max = 1), ncol = d)
  #   ridge.points = matrix(lowers, ncol = d, nrow = nrow(ridge.points),
  #                         byrow = TRUE) +
  #     ridge.points * (matrix(uppers, ncol = d, nrow = nrow(ridge.points), byrow = TRUE) -
  #                       matrix(lowers, ncol = d, nrow = nrow(ridge.points), byrow = TRUE))
  #
  # } else if (mesh == "LHS") {
  #   ridge.points = DiceDesign::lhsDesign(prod(mesh.size), dimension = d)$design
  #   ridge.points = matrix(lowers, ncol = d, nrow = nrow(ridge.points), byrow = TRUE) +
  #     ridge.points * (matrix(uppers, ncol = d, nrow = nrow(ridge.points), byrow = TRUE) -
  #                       matrix(lowers, ncol = d, nrow = nrow(ridge.points), byrow = TRUE))
  #
  # } else stop("unsupported mesh : ", mesh)
  #
  # b = c(lowers[1], uppers[1])
  # for (id in 1:(d - 1)) {
  #   b = rbind(cbind(b, lowers[id + 1]), cbind(b, uppers[id + 1]))
  # }
  # for (i in 1:nrow(b)) if (min_dist(b[i, ], ridge.points) > .Machine$double.eps)
  #   ridge.points = rbind(ridge.points, b[i, ])
  #
  # simplexes <- geometry::delaunayn(ridge.points, output.options = TRUE)

  simplexes = mesh(intervals, mesh.type, mesh.sizes)
  # warning(paste0("    [roots_mesh] mesh: ",Sys.time() - .t1))
  # .t1 <<- Sys.time()
  ridge.points = simplexes$p

  if (isTRUE(vectorized) && is.function(f)) {
    f_vec = function(x, ...) f(x, ...)
  } else if (is.function(vectorized)) {
    f_vec = function(x, ...) vectorized(x, ...)
    if (is.null(f)) f = f_vec
  } else if (isFALSE(vectorized) && !is.null(f)) {
    f_vec = Vectorize.function(f, dim = ncol(ridge.points))
  } else if (is.character(vectorized) && !is.null(f)) {
    f_vec = Vectorize.function(f, dim = ncol(ridge.points), .lapply = match.fun(vectorized))
  } else
    stop("Cannot decide how to vectorize f")

  ridge.y = f_vec(ridge.points, ...)
  # warning(paste0("    [roots_mesh] f_vec: ",Sys.time() - .t1))
  # .t1 <<- Sys.time()

  if (requireNamespace("arrangements"))
    tcombn2 = function(x) arrangements::combinations(x, 2)
  else tcombn2 = function(x) t(utils::combn(x, 2))

  ridges = NULL
  for (i in 1:nrow(simplexes$tri)) {
    more_ridges = tcombn2(simplexes$tri[i, ])
    for (j in 1:nrow(more_ridges))
      if (ridge.y[more_ridges[j,1]] * ridge.y[more_ridges[j, 2]] < 0)
        ridges = rbind(ridges, more_ridges[j, ])
  }
  # warning(paste0("    [roots_mesh] comb: ",Sys.time() - .t1))
  # .t1 <<- Sys.time()

  if (is.null(ridges)) {
    r = NA
    if (!is.null(ridge.points)) {
      attr(r, "mesh") <- simplexes
    }
    return(r)
  }

  ridges = unique(as.matrix(ridges))

  # r = Apply.function(.lapply = lapply,X = 1:length(ridges), MARGIN = 1, FUN = function(ridge_i) {
  #     X = ridge.points[ridge_i, , drop = FALSE]
  #     y = ridge.y[ridge_i]
  #     if (isFALSE((all(y > 0) || all(y < 0)))) {
  #         f.r = function(alpha, ...) {
  #           fa=as.numeric(f(alpha * X[1,] + (1 - alpha) * X[2, ], ...));
  #           # print(paste0("fa=",fa));
  #           fa}
  #         alpha_i = root(f.r, lower = 0, upper = 1, maxerror_f = maxerror_f,
  #                        tol = tol, f_lower = y[2], f_upper = y[1], ...)
  #         # print(paste0("a=",alpha_i))
  #         return(alpha_i * X[1, ,drop=FALSE] + (1 - alpha_i) * X[2, ,drop=FALSE])
  #     }
  #     else return(NULL)
  # }, .combine = rbind)

  root_fun = function(y,X,maxerror_f,tol) {
      if (isFALSE((all(y > 0) || all(y < 0)))) {
          f_r = function(alpha, ...) {
              as.numeric(f(alpha * X[1, ] + (1 - alpha) * X[2, ], ...))} # assumes one only root (inside current mesh element)
          alpha_i = 0 # default value if root finding fails after:
          try({alpha_i <- root(f=f_r, lower = 0, upper = 1, maxerror_f = maxerror_f,
                               tol = tol, f_lower = y[2], f_upper = y[1], max.rec=10, ...)})
          #return(alpha_i * X[1, ,drop=FALSE] + (1 - alpha_i) * X[2, ,drop=FALSE])
          # benchmark:
          #X = matrix(runif(10),ncol=5)
          #alpha_i = 0.25
          #rbenchmark::benchmark({c(alpha_i,1-alpha_i)%*%X},{alpha_i * X[1, ,drop=FALSE] + (1 - alpha_i) * X[2, ,drop=FALSE]},replications=100000)
          return( c(alpha_i,1-alpha_i) %*% X )
      } else return(NULL)
  }

  if (num_workers>1)
      r = do.call(rbind,parallel::mclapply(X = 1:nrow(ridges), FUN = function(ridge_i,...) {
        root_fun( y = ridge.y[ridges[ridge_i,]],
                  X = ridge.points[ridges[ridge_i,], , drop = FALSE],
                  maxerror_f,tol)
      }, mc.cores=num_workers))
  else
      r = do.call(rbind,lapply(X = 1:nrow(ridges), FUN = function(ridge_i,...) {
          root_fun( y = ridge.y[ridges[ridge_i,]],
                    X = ridge.points[ridges[ridge_i,], , drop = FALSE],
                    maxerror_f,tol)
      }))
  # r = apply(matrix(1:nrow(ridges)), 1, function(ridge_i) {
  #         root_fun( y = ridge.y[ridges[ridge_i,]],
  #                   X = ridge.points[ridges[ridge_i,], , drop = FALSE],
  #                   maxerror_f,tol)})

  # warning(paste0("    [roots_mesh] ridges root_fun: ",Sys.time() - .t1))
  # .t1 <<- Sys.time()

  if (is.null(r) || length(r) == 0)
    r = NA
  else {
    colnames(r) <- colnames(intervals)
    rownames(r) <- NULL
  }
  if (!is.null(ridge.points)) {
    attr(r, "mesh") <- simplexes
  }

  return(r)
}

#' @title Mesh level set of function
#' @param f function to be evaluated on the mesh
#' @param vectorized logical or function. If TRUE, f is assumed to be vectorized.
#' @param level level/threshold value
#' @param intervals matrix of intervals
#' @param mesh mesh object or type
#' @param ... additional arguments passed to f
#' @export
mesh_level = function(f, vectorized=FALSE, level=0, intervals, mesh, ...) {
  if (!is.mesh(mesh))
    mesh = mesh(intervals, mesh.type=mesh)

  if (is.array(f) && length(f)==nrow(mesh$p))
    mesh$y = f
  else {
    if (isTRUE(vectorized))
      f_vec=function(x,...) f(x,...)
    else if (is.function(vectorized))
      f_vec=function(x,...) vectorized(x,...)
    else
      f_vec=Vectorize.function(f,dim=ncol(mesh$p))
    mesh$y = f_vec(mesh$p,...)
  }
  ## Benchmarking
  # a = runif(5)-0.5
  # microbenchmark::microbenchmark(
  #   any(a>=0) && any(a<=0),
  #   any(diff(sign(a))!=0),
  #   times=100
  # )
  I = which(apply(mesh$tri,1,function(i) {yi=mesh$y[i]-level; any(yi>=0) && any(yi<=0)}))

  colnames(mesh$p) <- colnames(intervals)

  return(list(p=mesh$p,
              y=mesh$y,
              tri=mesh$tri[I,],
              areas=mesh$areas[I],
              neighbours=mesh$neighbours[I]))
}

#' @title Apply Functions Over Array Margins, using custom vectorization (possibly using parallel)
#' @description Emulate parallel apply on a function, from mclapply. Returns a vector or array or list of values obtained by applying a function to margins of an array or matrix.
#' @param FUN function to apply on X
#' @param X array of input values for FUN
#' @param MARGIN 1 indicates to apply on rows (default), 2 on columns
#' @param .combine how to combine results (default using c(.))
#' @param .lapply how to vectorize FUN call (default is parallel::mclapply)
#' @param ... optional arguments to FUN.
#' @return array of values taken by FUN on each row/column of X
#' @export
#' @examples
#' X = matrix(runif(10),ncol=2);
#'   rowSums(X) == apply(X,1,sum)
#'   apply(X,1,sum) == Apply.function(sum,X)
#'
#' X = matrix(runif(10),ncol=1)
#'   rowSums(X) == apply(X,1,sum)
#'   apply(X,1,sum) == Apply.function(sum,X)
#'
#' X = matrix(runif(10),ncol=2)
#' f = function(X) X[1]/X[2]
#' apply(X,1,f) == Apply.function(f,X)
Apply.function <- function(FUN, X, MARGIN=1, .combine=c, .lapply=parallel::mclapply, ...) {
    if (MARGIN==2) return(FUN, X=t(Apply.function(t(X), MARGIN=1, .combine=.combine, .lapply=.lapply,...)))
    if (MARGIN != 1) stop("Do not (yet) support MARGIN != 1")

    X.list = lapply(seq_len(nrow(X)), function(i) X[i,])
    l = .lapply(X.list,FUN,...)

    if (is.null(.combine)) return(l)
    do.call(.combine,l)
}

#' @title Vectorize a multidimensional Function
#' @description Vectorize a d-dimensional (input) function, in the same way that base::Vectorize for 1-dimensional functions.
#' @param fun 'dim'-dimensional function to Vectorize
#' @param dim dimension of input arguments of fun
#' @param .combine how to combine results (default using c(.))
#' @param .lapply how to vectorize FUN call (default is parallel::mclapply)
#' @param ... optional args to pass to 'Apply.function()', including .combine, .lapply, or optional args passed to 'fun'.
#' @return a vectorized function (to be called on matrix argument, on each row)
#' @export
#' @examples
#' f = function(x)x[1]+1; f(1:10); F = Vectorize.function(f,1);
#' F(1:10); #F = Vectorize(f); F(1:10);
#'
#' f2 = function(x)x[1]+x[2]; f2(1:10); F2 = Vectorize.function(f2,2);
#' F2(cbind(1:10,11:20));
#'
#' f3 = function(x)list(mean=x[1]+x[2],se=x[1]*x[2]); f3(1:10); F3 = Vectorize.function(f3,2);
#' F3(cbind(1:10,11:20));
Vectorize.function = function(fun, dim, .combine=rbind, .lapply=parallel::mclapply, ...) {
    function(X,...) {
        if (!is.matrix(X)) X = matrix(X,ncol=dim)
        Apply.function(fun,X, .combine=.combine, .lapply=.lapply,...)
    }
}


#' @title Memoize a function
#' @description Before each call of a function, check that the cache holds the results and returns it if available. Otherwise, compute f and cache the result for next evluations.
#' @param fun function to memoize
#' @param suffix suffix to use for cache files (default ".RcacheDiceView")
#' @return a function with same behavior than argument one, but using cache.
#' @importFrom R.cache loadCache
#' @importFrom R.cache saveCache
#' @export
#' @examples
#' f=function(n) rnorm(n);
#' F=Memoize.function(f);
#' F(5); F(6); F(5)
Memoize.function <- function(fun, suffix=".RcacheDiceView") {
    function(...) {
        arg = list(...)
        res <- loadCache(arg, suffix=suffix)
        if (!is.null(res)) {
            # cat("Loaded cached result\n")
            #cat(".")
            return(res)
        }
        # 2. If not available, generate it.
        # cat("Generating result from scratch...")
        #cat("o")
        res <- fun(...)
        # Emulate slow algorithm
        # cat("ok\n")
        saveCache(res, key=arg, comment="DiceView", suffix=suffix)
        res
    }
}

#' Create a Data Frame from all combinations of factor variables
#' @description Generalization of base::expand.grid to more than 2 variables.
#' @param d number of variables (taken in following arguments with modulo)
#' @param ... variables to combine, as arrays of values
#' @return data frame of all possible combinations of variables values
#' @export
#' @examples
#' expand.grids(d=1)
#' expand.grids(d=1,seq(f=0,t=1,l=11))
#' expand.grids(d=2)
#' expand.grids(d=2,seq(f=0,t=1,l=11))
#' expand.grids(d=2,seq(f=0,t=1,l=11),seq(0,1,l=3))
#' expand.grids(d=3,seq(f=0,t=1,l=5))
#' expand.grids(d=NULL,seq(f=0,t=1,l=5),seq(f=0,t=1,l=5),seq(f=0,t=1,l=5))
#' expand.grids(seq(f=0,t=1,l=5),seq(f=0,t=1,l=5),seq(f=0,t=1,l=5))
#' expand.grids(d=4,seq(f=0,t=1,l=5))
expand.grids <- function(d=length(list(...)),...) {
    if (isTRUE(is.na(d)) || isTRUE(is.null(d))) d = length(list(...)) # when d is explicitely NA, compute it.
    s = list(...)
    if (length(d)>0) { # When d is not set, it effectively contains first sample.
        s = c("d",s)
        s[[1]] = d
        d = length(s)
    }
    if (length(s)==0) s=list(c(0,1)) # when no ... provided, uses {0,1}
    if (length(s)==1 & is.list(s[[1]])) s = s[[1]] # when ... is already a list of samples
    l=list(s[[1]])
    if (d>1) for (i in 2:d) {
        l=c(l,list(s[[ (i-1) %% length(s) + 1 ]])) # weird formula when table index starts from 1...
    }
    return(do.call("expand.grid",l))
}

#' eval function and cast result to a list of y, y_low, y_up (possibly NA)
#' @param fun function to evaluate
#' @param X matrix of input values for fun
#' @param vectorized whether fun is vectorized or not
#' @param dim dimension of input values for fun if
#' @return list of y, y_low, y_up
EvalInterval.function = function(fun,X, vectorized=FALSE, dim=ncol(X)) {
        if (!vectorized)
            Fun = Vectorize.function(fun, dim)
        else
            Fun = fun

        if (is.null(X)) {
            if (is.null(dim)) dim = 1
            X = matrix(c(0,1), ncol=dim, nrow=2)
        }

        f_x <- Fun(as.matrix(X))

        # Now cast that eval :
        y <- NA #array(NA, npoints)
        y_low <- NA #array(NA, npoints)
        y_up <- NA #array(NA, npoints)
        if (is.list(f_x)) {
            f_x = lapply(as.list(as.data.frame(f_x)),unlist)
            if (!("mean" %in% names(f_x)) || !("se" %in% names(f_x)))
                stop(paste0("If function returns a list, it must have 'mean' and 'se', while was ",paste0(collapse="\n",utils::capture.output(print(f_x)))))
            y <- as.numeric(f_x$mean)
            if (!is.numeric(y))
                stop("If function returns a list, 'mean' must be (as) numeric:",paste0(y,collapse="\n"))

            if (!is.numeric(as.numeric(f_x$se)))
                stop("If function returns a list, 'se' must be (as) numeric:",paste0(f_x$se,collapse="\n"))
            y_low <- y - as.numeric(f_x$se)
            y_up <- y + as.numeric(f_x$se)

        } else if (is.matrix(f_x) && ncol(f_x)==2) {
            y_low <- as.numeric(f_x[,1])
            y_up <- as.numeric(f_x[,2])
            if (!is.numeric(y_low))
                stop("If function returns a matrix, first column must be (as) numeric.")
            if (!is.numeric(y_up))
                stop("If function returns a matrix, second column must be (as) numeric.")
        } else { # simple function, not a list nor a matrix
            y <- as.numeric(f_x)
            if (!is.numeric(y))
                stop("If function does not returns a list, it must be (as) numeric.")
        }

        return(list(y=y,y_low=y_low,y_up=y_up))
}

#### Excursion sets ####

#' @title Search excursion set of nD function, sampled by a mesh
#' @param f Function to inverse at 'threshold'
#' @param threshold target value to inverse
#' @param sign focus at conservative for above (sign=1) or below (sign=-1) the threshold
#' @param intervals bounds to inverse in, each column contains min and max of each dimension
#' @param mesh.type "unif" or "seq" (default) or "LHS" to preform interval partition
#' @param mesh.sizes number of parts for mesh (duplicate for each dimension if using "seq")
#' @param maxerror_f maximal tolerance on f precision
#' @param ex_filter.tri boolean function to validate a geometry::tri as considered in excursion : 'any' or 'all'
#' @param ... parameters to forward to roots_mesh(...) call
#' @param vectorized boolean: is f already vectorized ? (default: FALSE) or if function: vectorized version of f.
#' @param tol the desired accuracy (convergence tolerance on f arg).
#' @param num_workers number of cores to use for parallelization
#' @importFrom geometry delaunayn
#' @importFrom utils combn
#' @export
#' @examples
#' # mesh_exsets(function(x) x, threshold=.51, sign=1, intervals=rbind(0,1),
#' #   maxerror_f=1E-2,tol=1E-2, num_workers=1) # for faster testing
#' # mesh_exsets(function(x) x, threshold=.50000001, sign=1, intervals=rbind(0,1),
#' #   maxerror_f=1E-2,tol=1E-2, num_workers=1) # for faster testing
#' # mesh_exsets(function(x) sum(x), threshold=.51,sign=1, intervals=cbind(rbind(0,1),rbind(0,1)),
#' #   maxerror_f=1E-2,tol=1E-2, num_workers=1) # for faster testing
#' # mesh_exsets(sin,threshold=0,sign="sup",interval=c(pi/2,5*pi/2),
#' #   maxerror_f=1E-2,tol=1E-2, num_workers=1) # for faster testing
#'
#' if (identical(Sys.getenv("NOT_CRAN"), "true")) { # too long for CRAN on Windows
#'
#'   e = mesh_exsets(function(x) (0.25+x[1])^2+(0.5+x[2])^2 ,
#'                 threshold =0.25,sign=-1, intervals=matrix(c(-1,1,-1,1),nrow=2),
#'                 maxerror_f=1E-2,tol=1E-2, # for faster testing
#'                 num_workers=1)
#'
#'   plot(e$p,xlim=c(-1,1),ylim=c(-1,1));
#'   apply(e$tri,1,function(tri) polygon(e$p[tri,],col=rgb(.4,.4,.4,.4)))
#'   apply(e$frontiers,1,function(front) lines(e$p[front,],col='red'))
#'
#'   if (requireNamespace("rgl")) {
#'     e = mesh_exsets(function(x) (0.5+x[1])^2+(-0.5+x[2])^2+(0.+x[3])^2,
#'                   threshold = .25,sign=-1, mesh.type="unif",
#'                   intervals=matrix(c(-1,1,-1,1,-1,1),nrow=2),
#'                   maxerror_f=1E-2,tol=1E-2, # for faster testing
#'                   num_workers=1)
#'
#'     rgl::plot3d(e$p,xlim=c(-1,1),ylim=c(-1,1),zlim=c(-1,1));
#'     apply(e$tri,1,function(tri)rgl::lines3d(e$p[tri,]))
#'   }
#' }
mesh_exsets = function (f, vectorized = FALSE, threshold, sign, intervals,
      mesh.type = "seq", mesh.sizes = 11, maxerror_f = 1e-09, tol = .Machine$double.eps^0.25,
      ex_filter.tri = all, num_workers=maxWorkers(), ...) {
    # .t0 <<- Sys.time()

    if (sign == "lower" || sign == -1 || sign == "inf" || sign == "<" || isFALSE(sign))
        return(
            mesh_exsets(f = function(...) { -f(...) },
                        vectorized = vectorized, threshold = -threshold, sign = 1,
                        intervals = intervals, mesh.type = mesh.type, mesh.sizes = mesh.sizes,
                        maxerror_f = maxerror_f, tol = tol, ex_filter.tri=ex_filter.tri,
                        num_workers=num_workers, ...))

    if (sign != "upper" && sign != 1 && sign != "sup" && sign != ">" && !isTRUE(sign))
        stop("unknown sign: '", sign, "'")

    if (is.matrix(mesh.type))
        d = ncol(mesh.type)
    else if (is.matrix(intervals))
        d = ncol(intervals)
    else if (is.array(intervals))
        d = 1
    else
        stop("cannot identify dim of mesh")

    if (isTRUE(vectorized) && is.function(f)) {
        f_vec = function(x, ...) f(x, ...)
    } else if (is.function(vectorized)) {
        f_vec = function(x, ...) vectorized(x, ...)
        if (is.null(f)) f = f_vec
    } else if (isFALSE(vectorized) && !is.null(f)) {
        f_vec = Vectorize.function(f, dim = d)
    } else if (is.character(vectorized) && !is.null(f)) { # arbitrary lapply-like function
        f_vec = Vectorize.function(f, dim = d, .lapply = match.fun(vectorized))
    } else
        stop("Cannot decide how to vectorize f")

    # warning(paste0("[mesh_exsets] init: ",Sys.time() - .t0))
    # .t0 <<- Sys.time()

    k0 = 0#-1
    f_0 <- function(...) return(f(...) - (threshold + k0*maxerror_f))
    f_vec_0 <- function(...) return(f_vec(...) - (threshold + k0*maxerror_f))
    r <- roots_mesh(f = f_0, vectorized = f_vec_0, intervals = intervals,
                     mesh.type = mesh.type, mesh.sizes = mesh.sizes, maxerror_f = maxerror_f,
                     tol = tol, num_workers=num_workers, ...)

    # warning(paste0("[mesh_exsets] roots_mesh: ",Sys.time() - .t0))
    # .t0 <<- Sys.time()

    all_points = attr(r, "mesh")$p
    if (!all(is.na(r)))
        all_points = rbind(all_points, r)

    new_mesh <- mesh(intervals, mesh.type = all_points)

    # warning(paste0("[mesh_exsets] mesh: ",Sys.time() - .t0))
    # .t0 <<- Sys.time()

    # identify points belonging to the excursion set
    new_mesh$y = f_vec(new_mesh$p, ...) # re-evaluate f to get exact values (ie. close to frontier are 0.0 +/- maxrror_f)

    # warning(paste0("[mesh_exsets] f_vec: ",Sys.time() - .t0))
    # .t0 <<- Sys.time()

    I = which(apply(new_mesh$tri, 1,
                    function(i) ex_filter.tri(new_mesh$y[i] >= (threshold - 2*maxerror_f))))

    # warning(paste0("[mesh_exsets] I: ",Sys.time() - .t0))
    # .t0 <<- Sys.time()

    # identify edges at the frontier: one point of the tri is outside, the other inside
    d = ncol(new_mesh$p)
    frontiers = matrix(NA, nrow = 0, ncol = d)
    for (i in 1:nrow(new_mesh$tri)) {
        tri = new_mesh$tri[i,]
        outs = tri[ which(new_mesh$y[tri] <  (threshold - 2*maxerror_f))]
        ins =  tri[ which(new_mesh$y[tri] >= (threshold - 2*maxerror_f))]
        if (length(outs) == 1) {
            if (d==1)
                frontiers = rbind(frontiers, ins)
            else
                frontiers = rbind(frontiers, t(combn(ins, d)))
        }
    }
    frontiers = unique(frontiers)

    colnames(new_mesh$p) <- colnames(intervals)

    return(list(p = new_mesh$p, y = new_mesh$y,
                tri = new_mesh$tri[I, ],
                all_tri = new_mesh$tri,
                areas = new_mesh$areas[I],
                all_areas = new_mesh$areas,
                neighbours = new_mesh$neighbours[I],
                frontiers = frontiers))
}

#### Plot meshes ####

#' @title Plot a one dimensional mesh
#' @param mesh 1-dimensional mesh to draw
#' @param y ordinate value where to draw the mesh
#' @param color.nodes color of the mesh nodes
#' @param color.mesh color of the mesh elements
#' @param alpha transparency of the mesh elements & nodes
#' @param ... optional arguments passed to plot function
#' @import graphics
#' @export
#' @examples
#' plot_mesh(mesh_exsets(function(x) x, threshold=.51, sign=1,
#' intervals=rbind(0,1), num_workers=1))
#' plot_mesh(mesh_exsets(function(x) (x-.5)^2, threshold=.1, sign=-1,
#' intervals=rbind(0,1), num_workers=1))
plot_mesh = function(mesh,y=0,color.nodes='black',color.mesh='darkgray',alpha=0.4, ...){
    nodes.rgb=col2rgb(color.nodes)/255
    p = plot(x=mesh$p,y=rep(y,length(mesh$p)),ylab="",col=rgb(nodes.rgb[1,],nodes.rgb[2,],nodes.rgb[3,],alpha),...)
    mesh.rgb=col2rgb(color.mesh)/255
    apply(mesh$tri,1,
          function(tri) {
              lines(mesh$p[tri,],y=rep(0,length(tri)),col=rgb(mesh.rgb[1,],mesh.rgb[2,],mesh.rgb[3,],alpha))
              polygon(x=c(mesh$p[tri,],rev(mesh$p[tri,])),y=c(-1,-1,1,1),border = NA,col=rgb(mesh.rgb[1,],mesh.rgb[2,],mesh.rgb[3,],alpha))
          })
    invisible(p)
}

#' @title Plot a two dimensional mesh
#' @param mesh 2-dimensional mesh to draw
#' @param color.nodes color of the mesh nodes
#' @param color.mesh color of the mesh elements
#' @param alpha transparency of the mesh elements & nodes
#' @param ... optional arguments passed to plot function
#' @import graphics
#' @export
#' @examples
#' plot2d_mesh(mesh_exsets(f = function(x) sin(pi*x[1])*sin(pi*x[2]),
#'                         threshold=0,sign=1, mesh.type="unif",mesh.size=11,
#'                         intervals = matrix(c(1/2,5/2,1/2,5/2),nrow=2),
#'                         num_workers=1))
plot2d_mesh = function(mesh,color.nodes='black',color.mesh='darkgray',alpha=0.4,...){
    nodes.rgb=col2rgb(color.nodes)/255
    p = plot(mesh$p,col=rgb(nodes.rgb[1,],nodes.rgb[2,],nodes.rgb[3,],alpha),...)
    mesh.rgb=col2rgb(color.mesh)/255
    apply(mesh$tri,1,
          function(tri)
              polygon(mesh$p[c(tri,tri[1]),],border = rgb(mesh.rgb[1,],mesh.rgb[2,],mesh.rgb[3,],alpha), col = rgb(mesh.rgb[1,],mesh.rgb[2,],mesh.rgb[3,],alpha)))
    # sapply(1:nrow(mesh$tri),function(i) {xy=colMeans(mesh$p[mesh$tri[i,],]);text(x = xy[1],y=xy[2],paste0(i))})
    # apply(mesh$tri,1,function(tri) points(mesh$p[tri,], col = rgb(0,0,0,1),pch=20))
    invisible(p)
}

#' @title Plot a three dimensional mesh
#' @param mesh 3-dimensional mesh to draw
#' @param engine3d 3d framework to use: 'rgl' if installed or 'scatterplot3d' (default)
#' @param color.nodes color of the mesh nodes
#' @param color.mesh color of the mesh elements
#' @param alpha transparency of the mesh elements & nodes
#' @param ... optional arguments passed to plot function
#' @export
#' @examples
#' if (identical(Sys.getenv("NOT_CRAN"), "true")) { # too long for CRAN on Windows
#'
#'   plot3d_mesh(mesh_exsets(function(x) (0.5+x[1])^2+(-0.5+x[2])^2+(0.+x[3])^2,
#'                           threshold = .25,sign=-1, mesh.type="unif",
#'                           maxerror_f=1E-2,tol=1E-2, # faster display
#'                           intervals=matrix(c(-1,1,-1,1,-1,1),nrow=2),
#'                           num_workers=1),
#'                           engine3d='scatterplot3d')
#'
#'   if (requireNamespace("rgl")) {
#'     plot3d_mesh(mesh_exsets(function(x) (0.5+x[1])^2+(-0.5+x[2])^2+(0.+x[3])^2,
#'                             threshold = .25,sign=-1, mesh.type="unif",
#'                             maxerror_f=1E-2,tol=1E-2, # faster display
#'                             intervals=matrix(c(-1,1,-1,1,-1,1),nrow=2),
#'                             num_workers=1),engine3d='rgl')
#'   }
#' }
plot3d_mesh = function(mesh,engine3d=NULL,color.nodes='black',color.mesh='darkgray',alpha=0.4,...){
    nodes.rgb=col2rgb(color.nodes)/255
    package = .load3d(engine3d)
    if (is.null(package)) return()
    p3d = .plot3d(mesh$p,col=rgb(nodes.rgb[1,],nodes.rgb[2,],nodes.rgb[3,],alpha), package=package,...)
    mesh.rgb=col2rgb(color.mesh)/255
    apply(mesh$tri,1,function(tri) {
        .quads3d(mesh$p[tri,],col=rgb(mesh.rgb[1,],mesh.rgb[2,],mesh.rgb[3,]),alpha=alpha/10, package=package)
        .quads3d(mesh$p[tri,][c(4,3,2,1),],col=rgb(mesh.rgb[1,],mesh.rgb[2,],mesh.rgb[3,]),alpha=alpha/10, package=package)
        # .triangles3d(mesh$p[tri,][-1,],col=color,alpha=0.05, package=package)
        # .triangles3d(mesh$p[tri,][-2,],col=color,alpha=0.05, package=package)
        # .triangles3d(mesh$p[tri,][-3,],col=color,alpha=0.05, package=package)
        # .triangles3d(mesh$p[tri,][-4,],col=color,alpha=0.05, package=package)
    }) #rgl::lines3d(mesh$p[t(combn(tri,2)),],col=color))
    invisible(p3d)
}


#### Utility functions ####

#' @title Generalize expand.grid() for multi-columns data. Build all combinations of lines from X1 and X2. Each line may hold multiple columns.
#' @param X1 variable values, possibly with many columns
#' @param X2 variable values, possibly with many columns
#' combn.design(matrix(c(10,20),ncol=1),matrix(c(1,2,3,4,5,6),ncol=2))
#' combn.design(matrix(c(10,20,30,40),ncol=2),matrix(c(1,2,3,4,5,6),ncol=2))
combn.design <- function(X1,X2) {
    n1 = nrow(X1)
    n2 = nrow(X2)
    n = expand.grid(1:n1,1:n2)
    C = matrix(NA,ncol=ncol(X1)+ncol(X2),nrow=n1*n2)
    for (i in 1:(n1*n2)) {
        C[i,1:ncol(X1)] = X1[n[i,1],]
        C[i,ncol(X1)+(1:ncol(X2))] = X2[n[i,2],]
    }
    C
}

#' @title Test if points are in a hull
#' @param x points to test
#' @param p points defining the hull
#' @param h hull itself (built from p if given as NULL (default))
#' @export
#' @importFrom geometry convhulln
#' @importFrom geometry inhulln
#' @examples
#' is_in.p(x=-0.5,p=matrix(c(0,1),ncol=1))
#' is_in.p(x=0.5,p=matrix(c(0,1),ncol=1))
#' is_in.p(x=matrix(-.5,ncol=2,nrow=1),p=matrix(c(0,0,1,1,0,0),ncol=2))
#' is_in.p(x=matrix(.25,ncol=2,nrow=1),p=matrix(c(0,0,1,1,0,0),ncol=2))
#' is_in.p(x=matrix(-.5,ncol=3,nrow=1),p=matrix(c(0,0,0,1,0,0,0,1,0,0,0,1),ncol=3,byrow = TRUE))
#' is_in.p(x=matrix(.25,ncol=3,nrow=1),p=matrix(c(0,0,0,1,0,0,0,1,0,0,0,1),ncol=3,byrow = TRUE))
is_in.p = function(x,p,h=NULL) {
    if (is.null(h)) { # use p points, build h hull
        if (ncol(p)==1) return(x>min(p) && x<max(p))
        if (!is.matrix(x)) x = matrix(x,ncol=ncol(p))
        h = geometry::convhulln(p)
    }
    geometry::inhulln(ch = h,p=x)
}

#' Checks if some point belongs to a given mesh
#' @param x point to check
#' @param mesh mesh identifying the set which X may belong
#' @import stats
#' @export
#' @examples
#' is_in.mesh(-0.5,mesh=geometry::delaunayn(matrix(c(0,1),ncol=1),output.options =TRUE))
#' is_in.mesh(0.5,mesh=geometry::delaunayn(matrix(c(0,1),ncol=1),output.options =TRUE))
#'
#' x =matrix(-.5,ncol=2,nrow=1)
#' is_in.mesh(x,mesh=geometry::delaunayn(matrix(c(0,0,1,1,0,0),ncol=2),output.options =TRUE))
#'
#' x =matrix(.5,ncol=2,nrow=1)
#' is_in.mesh(x,mesh=geometry::delaunayn(matrix(c(0,0,1,1,0,0),ncol=2),output.options =TRUE))
is_in.mesh = function(x,mesh) {
    for (i in 1:nrow(mesh$tri)) {
        # if (mesh$areas[i]>.Machine$double.eps) # otherwise, flat triangle NOT RELIABLE
            if (all(dist(mesh$p[mesh$tri[i,],, drop = FALSE])>.Machine$double.eps)) {# not too close points
                #if (qr(mesh$p[mesh$tri[i,],, drop = FALSE])$rank == ncol(mesh$p)) # check this is not a degenerated (flat) triangle
                isin = FALSE
                try(isin <- is_in.p(x, mesh$p[mesh$tri[i,],, drop = FALSE]),silent = T) # to avoid coplanar errors in qhull
                if (isin) return(TRUE)
            } # else warning("Too flat element (",i,"): ",utils::capture.output(print(mesh$p[mesh$tri[i,],, drop = FALSE])))
        # else warning("Too small element (",i,"): area=",mesh$areas[i])
        }
    return(FALSE)
}

#' @title Checks if some points belong to a given mesh
#' @param X points to check
#' @param mesh mesh identifying the set which X may belong
#' @export
#' @examples
#' X = matrix(runif(100),ncol=2);
#' inside = are_in.mesh(X,mesh=geometry::delaunayn(matrix(c(0,0,1,1,0,0),ncol=2),output.options =TRUE))
#' print(inside)
#' plot(X,col=rgb(1-inside,0,0+inside))
are_in.mesh = function(X,mesh) {
    #apply(X,1,is_in.mesh,mesh)
    #unlist(foreach(i = 1:nrow(X)) %dopar% {is_in.mesh(X[i,],mesh)})
    X.list = lapply(seq_len(nrow(X)), function(i) X[i,])
    array(unlist(parallel::mclapply(X.list,is_in.mesh,mesh)))
}

#' @title Extract points of mesh which belong to the mesh triangulation (may not contain all points)
#' @param mesh mesh (list(p,tri,...) from geometry)
#' @return points coordinates inside the mesh triangulation
#' @export
points_in.mesh = function(mesh) {
    if (is.null(mesh)) return(NULL)
    mesh$p[unique(array(mesh$tri)),, drop = FALSE]
}
#' @title Extract points of mesh which do not belong to the mesh triangulation (may not contain all points)
#' @param mesh (list(p,tri,...) from geometry)
#' @return points coordinates outside the mesh triangulation
#' @export
points_out.mesh = function(mesh) {
    if (is.null(mesh)) return(NULL)
    mesh$p[-unique(array(mesh$tri)),, drop = FALSE]
}

#' @title Checks if a mesh is valid
#' @param x mesh to check
#' @export
#' @return TRUE if mesh is valid
is.mesh = function(x) {
    is.list(x) && all(c("p","tri") %in% names(x))
}


#' @title Compute distance between a point and a mesh
#' @param p point to compute distance from
#' @param mesh mesh to compute distance to
#' @param norm vector of weights for each dimension (default: 1)
#' @return distance between x and mesh
#' @export
#' @examples
#'  x = matrix(0,ncol=2)
#'  m = list(p = matrix(c(0,1,1,0,1,1),ncol=2,byrow=TRUE), tri = matrix(c(1,2,3),nrow=1))
#'  plot2d_mesh(m)
#'  points(x)
#'  min = min_dist.mesh(x,m)
#'  lines(rbind(x,attr(min,"proj")),col='red')
#'
#'  m = mesh_exsets(function(x) (0.25+x[1])^2+(0.5+x[2]/2)^2, vec=FALSE,
#'                  1 ,1, intervals=rbind(cbind(0,0),cbind(1,1)), num_workers=1)
#'  plot2d_mesh(m)
#'  x = matrix(c(0.25,0.25),ncol=2)
#'  points(x)
#'  min = min_dist.mesh(x,m)
#'  lines(rbind(x,attr(min,"proj")),col='red')
min_dist.mesh = function(p,mesh, norm=rep(1,ncol(mesh$p))) {
    dist2_to_p = function(x) sum(((p-x)/norm)^2)
    dists = NULL
    projs = NULL
    dim = ncol(mesh$p)
    for (i in 1:nrow(mesh$tri)) {
        trip = mesh$p[mesh$tri[i,],,drop=F]
        # find min dist bw p and tri, as a weighted barycenter of all points in the tri
        o = optim(rep(1,dim+1), # not optimal: 1 df too much
                  function(x) dist2_to_p((x/sum(x)) %*% trip),
                  method="L-BFGS-B",
                  lower=rep(1e-5,dim+1),upper=rep(1,dim+1)) # 1e-5: avoid 0 for sum(x)!=0, but that also excludes extreme points
        # ...so also compare dist with the extreme points of the tri:
        o$par = o$par/sum(o$par)
        best_points = rbind((o$par/sum(o$par)) %*% trip, trip)
        best_dists = sqrt(apply(best_points,1,dist2_to_p))
        best_i = which.min(best_dists)

        dists = c(dists, best_dists[best_i])
        projs = rbind(projs, best_points[best_i,])
    }
    im = which.min(dists)
    m = dists[im]
    attr(m,"proj") <- projs[im,]
    return(m)
}


#### Build mesh (on hypercube) ####

#' @title Builds a mesh from a design aor set of points
#' @param intervals bounds to inverse in, each column contains min and max (or values) of each dimension
#' @param mesh.type function or "unif" or "seq" (default) or "LHS" to preform interval partition
#' @param mesh.sizes number of parts for mesh (duplicate for each dimension if using "seq")
#' @return delaunay mesh (list(p,tri,...) from geometry)
#' @export
#' @examples
#' mesh = mesh(intervals=matrix(c(0,1,0,1),ncol=2),mesh.type="unif",mesh.sizes=10)
#' plot2d_mesh(mesh)
mesh = function(intervals, mesh.type = "seq", mesh.sizes = 11) {
    # setup bounds & dim
    if (is.matrix(intervals)) {
        if (nrow(intervals)!=2 && ncol(intervals)==2)
            intervals = t(intervals)
        lowers = apply(intervals, 2, min)
        uppers = apply(intervals, 2, max)
        d = ncol(intervals)
    } else if (is.null(intervals) && is.matrix(mesh.type)) {
        lowers = apply(mesh.type, 2, min)
        uppers = apply(mesh.type, 2, max)
        d = ncol(mesh.type)
    } else
        d = 1

    if (length(mesh.sizes) != d)
        mesh.size = rep(mesh.sizes, d)

    if (is.matrix(mesh.type)) {
        ridge.points = mesh.type
    } else if (is.function(mesh.type)) {
        ridge.points = mesh.type(prod(mesh.size), lowers, uppers)
    } else if (mesh.type == "seq") {
        ridge.points = matrix(seq(f = lowers[1], to = uppers[1],
                                  length.out = mesh.size[1]), ncol = 1)
        for (i in 2:d) {
            ridge.points = combn.design(ridge.points, matrix(seq(f = lowers[i], to = uppers[i], length.out = mesh.size[i]), ncol = 1))
        }
    } else if (mesh.type == "unif") {
        ridge.points = matrix(runif(n = d * prod(mesh.size),
                                    min = 0, max = 1), ncol = d)
        ridge.points = matrix(lowers, ncol = d, nrow = nrow(ridge.points),
                              byrow = TRUE) +
            ridge.points * (matrix(uppers, ncol = d, nrow = nrow(ridge.points), byrow = TRUE) -
                                matrix(lowers, ncol = d, nrow = nrow(ridge.points), byrow = TRUE))

    } else if (mesh.type == "LHS") {
        ridge.points = DiceDesign::lhsDesign(prod(mesh.size), dimension = d)$design
        ridge.points = matrix(lowers, ncol = d, nrow = nrow(ridge.points), byrow = TRUE) +
            ridge.points * (matrix(uppers, ncol = d, nrow = nrow(ridge.points), byrow = TRUE) -
                                matrix(lowers, ncol = d, nrow = nrow(ridge.points), byrow = TRUE))

    } else stop("unsupported mesh setup : ", mesh.type)

    # add bounds
    b = rbind(lowers[1], uppers[1])
    if (d>1) for (id in 1:(d - 1)) { # efficient way for factorial design
        b = rbind(cbind(b, lowers[id + 1]),cbind(b, uppers[id + 1]))
    }
    for (i in 1:nrow(b)) if (min_dist(b[i, ,drop=FALSE], ridge.points) > .Machine$double.eps)
        ridge.points = rbind(ridge.points, b[i, ,drop=FALSE])

    return( geometry::delaunayn(ridge.points, output.options = "Fn Fa", options= "Qt Qc Qz QbB Qcc") )
}

#### Generalization of optim (multistart) ####

#' Title optim wrapper for early stopping criterion
#' @param par starting point for optim
#' @param fn objective function, like in optim().
#' @param gr gradient function, like in optim().
#' @param fn.stop early stopping criterion
#' @param fn.NaN replacement value of fn when returns NaN
#' @param control control parameters for optim()
#' @param ... additional arguments passed to optim()
#' @return list with best solution and all solutions
#' @export
#' @author Yann Richet, ASNR
#' @examples
#' fn = function(x) x^6
#' o = optim( par=15, fn,lower=-20,upper=20,method='L-BFGS-B')
#' o.s = optim.stop( par=15, fn,lower=-20,upper=20,method='L-BFGS-B',fn.stop=0.1)
#' #check o.s$value == 0.1 && o.s$counts < o$counts
optim.stop <- function(par, fn, gr=NULL,
                       fn.stop=NA, fn.NaN=NaN,
                       control=list(), ...) {
    if (is.na(fn.stop)) {
        fn.o = function(...) {
            y = fn(...)
            y[which(is.nan(y))] <- fn.NaN
            return(y)
        }
        gr.o = gr
    } else {
        fn.o = function(...) {
            y = fn(...)
            if ((isTRUE(fn.stop > y) && !isFALSE(control$fnscale > 0)) || (isTRUE(fn.stop < y) && isTRUE(control$fnscale < 0))) {# reach early stopping criterion
                warning(paste("Early stopping criterion reached at", paste0(round(y, 3), collapse = ","), "for", paste0(round(par, 3), collapse = ",")))
                return(fn.stop)
            }
            y[which(is.nan(y))] <- fn.NaN
            return(y)
        }
        gr.o = if (is.null(gr)) NULL else function(...) {
            dy = gr(...)
            y = fn(...)
            if ((isTRUE(fn.stop > y) && !isFALSE(control$fnscale > 0)) || (isTRUE(fn.stop < y) && isTRUE(control$fnscale < 0))) # reach early stopping criterion
                return(rep(0,length(dy)))
            return(dy)
        }

    }
    optim(par=par, fn=fn.o, gr=gr.o, control=control,...)
}

#' Title Multi-local optimization wrapper for optim, using (possibly parallel) multistart.
#'
#' @param pars starting points for optim
#' @param fn objective function, like in optim().
#' @param fn.NaN replacement value of fn when returns NaN
#' @param fn.stop early stopping criterion
#' @param .apply loop/parallelization backend for multistart ("mclapply", "lapply" or "foreach")
#' @param pars.eps minimal distance between two solutions to be considered different
#' @param control control parameters for optim()
#' @param ... additional arguments passed to optim()
#'
#' @return list with best solution and all solutions
#' @importFrom foreach foreach
#' @importFrom foreach %dopar%
#' @export
#' @author Yann Richet, ASNR
#' @examples
#' fn = function(x) ifelse(x==0,1,sin(x)/x)
#' # plot(fn, xlim=c(-20,20))
#' optim( par=5,                     fn,lower=-20,upper=20,method='L-BFGS-B')
#' optims(pars=t(t(seq(-20,20,,20))),fn,lower=-20,upper=20,method='L-BFGS-B')
#'
#' # Branin function (3 local minimas)
#' f = function (x) {
#'   x1 <- x[1] * 15 - 5
#'   x2 <- x[2] * 15
#'   (x2 - 5/(4 * pi^2) * (x1^2) + 5/pi * x1 - 6)^2 + 10 * (1 - 1/(8 * pi)) * cos(x1) + 10
#' }
#' # expect to find 3 local minimas
#' optims(pars=matrix(runif(100),ncol=2),f,method="L-BFGS-B",lower=c(0,0),upper=c(1,1))
optims <- function(pars,fn,fn.NaN=NaN,fn.stop=NA,.apply="mclapply",pars.eps=1E-5,control=list(),...) {
    if (is.character(.apply)) {
        if (.apply=="lapply")
            O = lapply(X=as.list(as.data.frame(t(pars))),
                       FUN=function(x0){
                           try(optim.stop(par=x0,fn=fn, fn.stop=fn.stop,fn.NaN=fn.NaN, control=control,...))})
        else if (.apply=="mclapply")
            O = parallel::mclapply(X=as.list(as.data.frame(t(pars))),
                                   FUN=function(x0){
                                       try(optim.stop(par=x0,fn=fn, fn.stop=fn.stop,fn.NaN=fn.NaN, control=control,...))})
        else if (.apply=="foreach") {
            i.o = NULL
            O = foreach::foreach(i.o = 1:nrow(pars),.errorhandling = "remove") %dopar% { return(
                try(optim.stop(par=pars[i.o,],fn=fn, fn.stop=fn.stop,fn.NaN=fn.NaN, control=control,...))) }
        }
    } else
        O = .apply(X=as.list(as.data.frame(t(pars))),
                   FUN=function(x0){return(
                       try(optim.stop(par=x0,fn=fn, fn.stop=fn.stop,fn.NaN=fn.NaN, control=control,...)))})

    best = list(par=NA,value=NA)
    all = NULL
    is_best = function(best_value,o_value) {
        if (is.numeric(best_value) && is.numeric(o_value)) {
            if (!is.null(control$fnscale) && control$fnscale < 0)
                return(!isTRUE(best_value > o_value))
            else
                return(!isTRUE(best_value < o_value))
        } else return(FALSE)
    }
    for (o in O) {if ("value" %in% names(o)) {
        if (is.numeric(o$value) && !is.numeric(best$value)) # first value
            best = o
        if (is.numeric(o$value) && is_best(best$value, o$value)) # best value
            best = o
        if (is.null(all))
            all = list(pars=matrix(o$par,nrow=1),values=matrix(o$value,nrow=1))
        else if (min_dist(o$par,all$pars)>pars.eps) {
            all$pars=rbind(all$pars,o$par)
            all$values=rbind(all$values,o$value)
        }
    }}
    c(best,all)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/function.R
\name{EvalInterval.function}
\alias{EvalInterval.function}
\title{eval function and cast result to a list of y, y_low, y_up (possibly NA)}
\usage{
EvalInterval.function(fun, X, vectorized = FALSE, dim = ncol(X))
}
\arguments{
\item{fun}{function to evaluate}

\item{X}{matrix of input values for fun}

\item{vectorized}{whether fun is vectorized or not}

\item{dim}{dimension of input values for fun if}
}
\value{
list of y, y_low, y_up
}
\description{
eval function and cast result to a list of y, y_low, y_up (possibly NA)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/mesh_set.R
\name{plot3d_mesh}
\alias{plot3d_mesh}
\title{Plot a three dimensional mesh}
\usage{
plot3d_mesh(
  mesh,
  engine3d = NULL,
  color.nodes = "black",
  color.mesh = "darkgray",
  alpha = 0.4,
  ...
)
}
\arguments{
\item{mesh}{3-dimensional mesh to draw}

\item{engine3d}{3d framework to use: 'rgl' if installed or 'scatterplot3d' (default)}

\item{color.nodes}{color of the mesh nodes}

\item{color.mesh}{color of the mesh elements}

\item{alpha}{transparency of the mesh elements & nodes}

\item{...}{optional arguments passed to plot function}
}
\description{
Plot a three dimensional mesh
}
\examples{
if (identical(Sys.getenv("NOT_CRAN"), "true")) { # too long for CRAN on Windows

  plot3d_mesh(mesh_exsets(function(x) (0.5+x[1])^2+(-0.5+x[2])^2+(0.+x[3])^2,
                          threshold = .25,sign=-1, mesh.type="unif",
                          maxerror_f=1E-2,tol=1E-2, # faster display
                          intervals=matrix(c(-1,1,-1,1,-1,1),nrow=2),
                          num_workers=1),
                          engine3d='scatterplot3d')

  if (requireNamespace("rgl")) {
    plot3d_mesh(mesh_exsets(function(x) (0.5+x[1])^2+(-0.5+x[2])^2+(0.+x[3])^2,
                            threshold = .25,sign=-1, mesh.type="unif",
                            maxerror_f=1E-2,tol=1E-2, # faster display
                            intervals=matrix(c(-1,1,-1,1,-1,1),nrow=2),
                            num_workers=1),engine3d='rgl')
  }
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/function.R
\name{Memoize.function}
\alias{Memoize.function}
\title{Memoize a function}
\usage{
Memoize.function(fun, suffix = ".RcacheDiceView")
}
\arguments{
\item{fun}{function to memoize}

\item{suffix}{suffix to use for cache files (default ".RcacheDiceView")}
}
\value{
a function with same behavior than argument one, but using cache.
}
\description{
Before each call of a function, check that the cache holds the results and returns it if available. Otherwise, compute f and cache the result for next evluations.
}
\examples{
f=function(n) rnorm(n);
F=Memoize.function(f);
F(5); F(6); F(5)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/mesh_set.R
\name{points_in.mesh}
\alias{points_in.mesh}
\title{Extract points of mesh which belong to the mesh triangulation (may not contain all points)}
\usage{
points_in.mesh(mesh)
}
\arguments{
\item{mesh}{mesh (list(p,tri,...) from geometry)}
}
\value{
points coordinates inside the mesh triangulation
}
\description{
Extract points of mesh which belong to the mesh triangulation (may not contain all points)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/mesh_set.R
\name{is_in.p}
\alias{is_in.p}
\title{Test if points are in a hull}
\usage{
is_in.p(x, p, h = NULL)
}
\arguments{
\item{x}{points to test}

\item{p}{points defining the hull}

\item{h}{hull itself (built from p if given as NULL (default))}
}
\description{
Test if points are in a hull
}
\examples{
is_in.p(x=-0.5,p=matrix(c(0,1),ncol=1))
is_in.p(x=0.5,p=matrix(c(0,1),ncol=1))
is_in.p(x=matrix(-.5,ncol=2,nrow=1),p=matrix(c(0,0,1,1,0,0),ncol=2))
is_in.p(x=matrix(.25,ncol=2,nrow=1),p=matrix(c(0,0,1,1,0,0),ncol=2))
is_in.p(x=matrix(-.5,ncol=3,nrow=1),p=matrix(c(0,0,0,1,0,0,0,1,0,0,0,1),ncol=3,byrow = TRUE))
is_in.p(x=matrix(.25,ncol=3,nrow=1),p=matrix(c(0,0,0,1,0,0,0,1,0,0,0,1),ncol=3,byrow = TRUE))
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/optim.R
\name{optim.stop}
\alias{optim.stop}
\title{Title optim wrapper for early stopping criterion}
\usage{
optim.stop(
  par,
  fn,
  gr = NULL,
  fn.stop = NA,
  fn.NaN = NaN,
  control = list(),
  ...
)
}
\arguments{
\item{par}{starting point for optim}

\item{fn}{objective function, like in optim().}

\item{gr}{gradient function, like in optim().}

\item{fn.stop}{early stopping criterion}

\item{fn.NaN}{replacement value of fn when returns NaN}

\item{control}{control parameters for optim()}

\item{...}{additional arguments passed to optim()}
}
\value{
list with best solution and all solutions
}
\description{
Title optim wrapper for early stopping criterion
}
\examples{
fn = function(x) x^6
o = optim( par=15, fn,lower=-20,upper=20,method='L-BFGS-B')
o.s = optim.stop( par=15, fn,lower=-20,upper=20,method='L-BFGS-B',fn.stop=0.1)
#check o.s$value == 0.1 && o.s$counts < o$counts
}
\author{
Yann Richet, ASNR
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/mesh_set.R
\name{plot_mesh}
\alias{plot_mesh}
\title{Plot a one dimensional mesh}
\usage{
plot_mesh(
  mesh,
  y = 0,
  color.nodes = "black",
  color.mesh = "darkgray",
  alpha = 0.4,
  ...
)
}
\arguments{
\item{mesh}{1-dimensional mesh to draw}

\item{y}{ordinate value where to draw the mesh}

\item{color.nodes}{color of the mesh nodes}

\item{color.mesh}{color of the mesh elements}

\item{alpha}{transparency of the mesh elements & nodes}

\item{...}{optional arguments passed to plot function}
}
\description{
Plot a one dimensional mesh
}
\examples{
plot_mesh(mesh_exsets(function(x) x, threshold=.51, sign=1,
intervals=rbind(0,1), num_workers=1))
plot_mesh(mesh_exsets(function(x) (x-.5)^2, threshold=.1, sign=-1,
intervals=rbind(0,1), num_workers=1))
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/root.R
\name{roots_mesh}
\alias{roots_mesh}
\title{Multi Dimensional Multiple Roots (Zero) Finding, sampled by a mesh}
\usage{
roots_mesh(
  f,
  vectorized = FALSE,
  intervals,
  mesh.type = "seq",
  mesh.sizes = 11,
  maxerror_f = 1e-07,
  tol = .Machine$double.eps^0.25,
  num_workers = maxWorkers(),
  ...
)
}
\arguments{
\item{f}{Function (one or more dimensions) to find roots of}

\item{vectorized}{vectorized f: function, TRUE (use f directly), or wrap in Vectorize.function: FALSE (default args), "lapply", "mclapply", ...}

\item{intervals}{bounds to inverse in, each column contains min and max of each dimension}

\item{mesh.type}{function or "unif" or "seq" (default) to preform interval partition}

\item{mesh.sizes}{number of parts for mesh (duplicate for each dimension if using "seq")}

\item{maxerror_f}{the maximum error on f evaluation (iterates over uniroot to converge).}

\item{tol}{the desired accuracy (convergence tolerance on f arg).}

\item{num_workers}{number of parallel roots finding}

\item{...}{Other args for f}
}
\value{
matrix of x, so f(x)=0
}
\description{
Multi Dimensional Multiple Roots (Zero) Finding, sampled by a mesh
}
\examples{
roots_mesh(function(x) x-.51, intervals=rbind(0,1),
  num_workers=1)
roots_mesh(function(x) sum(x)-.51, intervals=cbind(rbind(0,1),rbind(0,1)),
  num_workers=1)
roots_mesh(sin,intervals=c(pi/2,5*pi/2),
  num_workers=1)
roots_mesh(f = function(x) sin(pi*x[1])*sin(pi*x[2]),
           intervals = matrix(c(1/2,5/2,1/2,5/2),nrow=2),
           num_workers=1)

r = roots_mesh(f = function(x) (0.25+x[1])^2+(0.5+x[2])^2 - .25,
intervals=matrix(c(-1,1,-1,1),nrow=2), mesh.size=5,
  num_workers=1)
plot(r,xlim=c(-1,1),ylim=c(-1,1))

r = roots_mesh(function(x) (0.5+x[1])^2+(-0.5+x[2])^2+(0.+x[3])^2 - .5,
               mesh.sizes = 11,
               intervals=matrix(c(-1,1,-1,1,-1,1),nrow=2),
               num_workers=1)
scatterplot3d::scatterplot3d(r,xlim=c(-1,1),ylim=c(-1,1),zlim=c(-1,1))

roots_mesh(function(x)exp(x)-1,intervals=c(-1,2),
           num_workers=1)
roots_mesh(function(x)exp(1000*x)-1,intervals=c(-1,2),
           num_workers=1)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/root.R
\name{roots}
\alias{roots}
\title{One Dimensional Multiple Roots (Zero) Finding}
\usage{
roots(
  f,
  vectorized = FALSE,
  interval,
  maxerror_f = 1e-07,
  split = "seq",
  split.size = 11,
  tol = .Machine$double.eps^0.25,
  .lapply = parallel::mclapply,
  ...
)
}
\arguments{
\item{f}{Function to find roots}

\item{vectorized}{boolean: is f already vectorized ? (default: FALSE) or if function: vectorized version of f.}

\item{interval}{bounds to inverse in}

\item{maxerror_f}{the maximum error on f evaluation (iterates over uniroot to converge).}

\item{split}{function or "unif" or "seq" (default) to preform interval partition}

\item{split.size}{number of parts to perform uniroot inside}

\item{tol}{the desired accuracy (convergence tolerance on f arg).}

\item{.lapply}{control the loop/vectorization over different roots (defaults to multicore apply).}

\item{...}{additional named or unnamed arguments to be passed to f.}
}
\value{
array of x, so f(x)=target
}
\description{
Search multiple roots of 1D function, sampled/splitted by a (1D) mesh
}
\examples{
roots(sin,interval=c(pi/2,5*pi/2))
roots(sin,interval=c(pi/2,1.5*pi/2))

f=function(x)exp(x)-1;
f(roots(f,interval=c(-1,2)))

f=function(x)exp(1000*x)-1;
f(roots(f,interval=c(-1,2)))
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/function.R
\name{Vectorize.function}
\alias{Vectorize.function}
\title{Vectorize a multidimensional Function}
\usage{
Vectorize.function(
  fun,
  dim,
  .combine = rbind,
  .lapply = parallel::mclapply,
  ...
)
}
\arguments{
\item{fun}{'dim'-dimensional function to Vectorize}

\item{dim}{dimension of input arguments of fun}

\item{.combine}{how to combine results (default using c(.))}

\item{.lapply}{how to vectorize FUN call (default is parallel::mclapply)}

\item{...}{optional args to pass to 'Apply.function()', including .combine, .lapply, or optional args passed to 'fun'.}
}
\value{
a vectorized function (to be called on matrix argument, on each row)
}
\description{
Vectorize a d-dimensional (input) function, in the same way that base::Vectorize for 1-dimensional functions.
}
\examples{
f = function(x)x[1]+1; f(1:10); F = Vectorize.function(f,1);
F(1:10); #F = Vectorize(f); F(1:10);

f2 = function(x)x[1]+x[2]; f2(1:10); F2 = Vectorize.function(f2,2);
F2(cbind(1:10,11:20));

f3 = function(x)list(mean=x[1]+x[2],se=x[1]*x[2]); f3(1:10); F3 = Vectorize.function(f3,2);
F3(cbind(1:10,11:20));
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/mesh_set.R
\name{combn.design}
\alias{combn.design}
\title{Generalize expand.grid() for multi-columns data. Build all combinations of lines from X1 and X2. Each line may hold multiple columns.}
\usage{
combn.design(X1, X2)
}
\arguments{
\item{X1}{variable values, possibly with many columns}

\item{X2}{variable values, possibly with many columns
combn.design(matrix(c(10,20),ncol=1),matrix(c(1,2,3,4,5,6),ncol=2))
combn.design(matrix(c(10,20,30,40),ncol=2),matrix(c(1,2,3,4,5,6),ncol=2))}
}
\description{
Generalize expand.grid() for multi-columns data. Build all combinations of lines from X1 and X2. Each line may hold multiple columns.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/function.R
\name{expand.grids}
\alias{expand.grids}
\title{Create a Data Frame from all combinations of factor variables}
\usage{
expand.grids(d = length(list(...)), ...)
}
\arguments{
\item{d}{number of variables (taken in following arguments with modulo)}

\item{...}{variables to combine, as arrays of values}
}
\value{
data frame of all possible combinations of variables values
}
\description{
Generalization of base::expand.grid to more than 2 variables.
}
\examples{
expand.grids(d=1)
expand.grids(d=1,seq(f=0,t=1,l=11))
expand.grids(d=2)
expand.grids(d=2,seq(f=0,t=1,l=11))
expand.grids(d=2,seq(f=0,t=1,l=11),seq(0,1,l=3))
expand.grids(d=3,seq(f=0,t=1,l=5))
expand.grids(d=NULL,seq(f=0,t=1,l=5),seq(f=0,t=1,l=5),seq(f=0,t=1,l=5))
expand.grids(seq(f=0,t=1,l=5),seq(f=0,t=1,l=5),seq(f=0,t=1,l=5))
expand.grids(d=4,seq(f=0,t=1,l=5))
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/mesh_set.R
\name{are_in.mesh}
\alias{are_in.mesh}
\title{Checks if some points belong to a given mesh}
\usage{
are_in.mesh(X, mesh)
}
\arguments{
\item{X}{points to check}

\item{mesh}{mesh identifying the set which X may belong}
}
\description{
Checks if some points belong to a given mesh
}
\examples{
X = matrix(runif(100),ncol=2);
inside = are_in.mesh(X,mesh=geometry::delaunayn(matrix(c(0,0,1,1,0,0),ncol=2),output.options =TRUE))
print(inside)
plot(X,col=rgb(1-inside,0,0+inside))
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/root.R
\name{root}
\alias{root}
\title{One Dimensional Root (Zero) Finding}
\usage{
root(
  f,
  lower,
  upper,
  maxerror_f = 1e-07,
  f_lower = f(lower, ...),
  f_upper = f(upper, ...),
  tol = .Machine$double.eps^0.25,
  convexity = FALSE,
  rec = 0,
  max.rec = NA,
  ...
)
}
\arguments{
\item{f}{the function for which the root is sought.}

\item{lower}{the lower end point of the interval to be searched.}

\item{upper}{the upper end point of the interval to be searched.}

\item{maxerror_f}{the maximum error on f evaluation (iterates over uniroot to converge).}

\item{f_lower}{the same as f(lower).}

\item{f_upper}{the same as f(upper).}

\item{tol}{the desired accuracy (convergence tolerance on f arg).}

\item{convexity}{the learned convexity factor of the function, used to reduce the boundaries for uniroot.}

\item{rec}{counter of recursive level.}

\item{max.rec}{maximal number of recursive level before failure (stop).}

\item{...}{additional named or unnamed arguments to be passed to f.}
}
\description{
Search one root with given precision (on y). Iterate over uniroot as long as necessary.
}
\examples{
f=function(x) {cat("f");1-exp(x)}; f(root(f,lower=-1,upper=2))
f=function(x) {cat("f");exp(x)-1}; f(root(f,lower=-1,upper=2))

.f = function(x) 1-exp(1*x)
f=function(x) {cat("f");y=.f(x);points(x,y,pch=20,col=rgb(0,0,0,.2));y}
plot(.f,xlim=c(-1,2)); f(root(f,lower=-1,upper=2))

.f = function(x) exp(10*x)-1
f=function(x) {cat("f");y=.f(x);points(x,y,pch=20);y}
plot(.f,xlim=c(-1,2)); f(root(f,lower=-1,upper=2))

.f = function(x) exp(100*x)-1
f=function(x) {cat("f");y=.f(x);points(x,y,pch=20);y}
plot(.f,xlim=c(-1,2)); f(root(f,lower=-1,upper=2))

f=function(x) {cat("f");exp(100*x)-1}; f(root(f,lower=-1,upper=2))

\dontrun{

  # Quite hard functions to find roots

  ## Increasing function
  ## convex
  n.f=0
  .f = function(x) exp(10*x)-1
  f=function(x) {n.f<<-n.f+1;y=.f(x);points(x,y,pch=20);y}
  plot(.f,xlim=c(-.1,.2)); f(root(f,lower=-1,upper=2))
  print(n.f)
  ## non-convex
  n.f=0
  .f = function(x) 1-exp(-10*x)
  f=function(x) {n.f<<-n.f+1;y=.f(x);points(x,y,pch=20);y}
  plot(.f,xlim=c(-.1,.2)); f(root(f,lower=-1,upper=2))
  print(n.f)

  # ## Decreasing function
  # ## non-convex
  n.f=0
  .f = function(x) 1-exp(10*x)
  f=function(x) {n.f<<-n.f+1;y=.f(x);points(x,y,pch=20,col=rgb(0,0,0,.2));y}
  plot(.f,xlim=c(-.1,.2)); f(root(f,lower=-1,upper=2))
  print(n.f)
  # ## convex
  n.f=0
  .f = function(x) exp(-10*x)-1
  f=function(x) {n.f<<-n.f+1;y=.f(x);points(x,y,pch=20,col=rgb(0,0,0,.2));y}
  plot(.f,xlim=c(-.1,.2)); f(root(f,lower=-1,upper=2))
  print(n.f)
}
}
\author{
Yann Richet, ASNR
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/mesh_set.R
\name{min_dist.mesh}
\alias{min_dist.mesh}
\title{Compute distance between a point and a mesh}
\usage{
min_dist.mesh(p, mesh, norm = rep(1, ncol(mesh$p)))
}
\arguments{
\item{p}{point to compute distance from}

\item{mesh}{mesh to compute distance to}

\item{norm}{vector of weights for each dimension (default: 1)}
}
\value{
distance between x and mesh
}
\description{
Compute distance between a point and a mesh
}
\examples{
 x = matrix(0,ncol=2)
 m = list(p = matrix(c(0,1,1,0,1,1),ncol=2,byrow=TRUE), tri = matrix(c(1,2,3),nrow=1))
 plot2d_mesh(m)
 points(x)
 min = min_dist.mesh(x,m)
 lines(rbind(x,attr(min,"proj")),col='red')

 m = mesh_exsets(function(x) (0.25+x[1])^2+(0.5+x[2]/2)^2, vec=FALSE,
                 1 ,1, intervals=rbind(cbind(0,0),cbind(1,1)), num_workers=1)
 plot2d_mesh(m)
 x = matrix(c(0.25,0.25),ncol=2)
 points(x)
 min = min_dist.mesh(x,m)
 lines(rbind(x,attr(min,"proj")),col='red')
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/function.R
\name{Apply.function}
\alias{Apply.function}
\title{Apply Functions Over Array Margins, using custom vectorization (possibly using parallel)}
\usage{
Apply.function(
  FUN,
  X,
  MARGIN = 1,
  .combine = c,
  .lapply = parallel::mclapply,
  ...
)
}
\arguments{
\item{FUN}{function to apply on X}

\item{X}{array of input values for FUN}

\item{MARGIN}{1 indicates to apply on rows (default), 2 on columns}

\item{.combine}{how to combine results (default using c(.))}

\item{.lapply}{how to vectorize FUN call (default is parallel::mclapply)}

\item{...}{optional arguments to FUN.}
}
\value{
array of values taken by FUN on each row/column of X
}
\description{
Emulate parallel apply on a function, from mclapply. Returns a vector or array or list of values obtained by applying a function to margins of an array or matrix.
}
\examples{
X = matrix(runif(10),ncol=2);
  rowSums(X) == apply(X,1,sum)
  apply(X,1,sum) == Apply.function(sum,X)

X = matrix(runif(10),ncol=1)
  rowSums(X) == apply(X,1,sum)
  apply(X,1,sum) == Apply.function(sum,X)

X = matrix(runif(10),ncol=2)
f = function(X) X[1]/X[2]
apply(X,1,f) == Apply.function(f,X)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/mesh_set.R
\name{is.mesh}
\alias{is.mesh}
\title{Checks if a mesh is valid}
\usage{
is.mesh(x)
}
\arguments{
\item{x}{mesh to check}
}
\value{
TRUE if mesh is valid
}
\description{
Checks if a mesh is valid
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/sectionview3d.R
\name{sectionview3d.function}
\alias{sectionview3d.function}
\alias{sectionview3d,function,function-method}
\alias{sectionview3d.matrix}
\alias{sectionview3d,matrix,matrix-method}
\alias{sectionview3d.character}
\alias{sectionview3d,character,character-method}
\alias{sectionview3d.km}
\alias{sectionview3d,km,km-method}
\alias{sectionview3d.Kriging}
\alias{sectionview3d,Kriging,Kriging-method}
\alias{sectionview3d.WarpKriging}
\alias{sectionview3d,WarpKriging,WarpKriging-method}
\alias{sectionview3d.glm}
\alias{sectionview3d,glm,glm-method}
\alias{sectionview3d.list}
\alias{sectionview3d,list,list-method}
\alias{sectionview3d}
\title{Plot a contour view of a prediction model or function, including design points if available.}
\usage{
\method{sectionview3d}{`function`}(
  fun,
  vectorized = FALSE,
  center = NULL,
  axis = NULL,
  npoints = 21,
  col_fun = if (!is.null(col)) col else "blue",
  col = NULL,
  col_fading_interval = 0.5,
  mfrow = c(1, 1),
  Xlab = NULL,
  ylab = NULL,
  Xlim = if (!add) matrix(c(0, 1), 2, 2) else NULL,
  ylim = NULL,
  title = NULL,
  title_sep = " | ",
  add = FALSE,
  engine3d = NULL,
  ...
)

\method{sectionview3d}{matrix}(
  X,
  y,
  center = NULL,
  axis = NULL,
  col_points = if (!is.null(col)) col else "red",
  col = NULL,
  col_fading_interval = 0.5,
  bg_fading = 1,
  mfrow = c(1, 1),
  Xlab = NULL,
  ylab = NULL,
  Xlim = if (!add) matrix(c(0, 1), 2, 2) else NULL,
  ylim = NULL,
  title = NULL,
  title_sep = " | ",
  add = FALSE,
  engine3d = NULL,
  ...
)

\method{sectionview3d}{character}(eval_str, axis = NULL, mfrow = c(1, 1), ...)

\method{sectionview3d}{km}(
  km_model,
  type = "UK",
  center = NULL,
  axis = NULL,
  npoints = 21,
  col_points = if (!is.null(col)) col else "red",
  col_fun = if (!is.null(col)) col else "blue",
  col = NULL,
  conf_level = 0.95,
  conf_fading = 0.5,
  bg_fading = 1,
  mfrow = c(1, 1),
  Xlab = NULL,
  ylab = NULL,
  Xlim = NULL,
  ylim = NULL,
  title = NULL,
  title_sep = " | ",
  add = FALSE,
  engine3d = NULL,
  ...
)

\method{sectionview3d}{Kriging}(
  Kriging_model,
  center = NULL,
  axis = NULL,
  npoints = 21,
  col_points = if (!is.null(col)) col else "red",
  col_fun = if (!is.null(col)) col else "blue",
  col = NULL,
  conf_level = 0.95,
  conf_fading = 0.5,
  bg_fading = 1,
  mfrow = c(1, 1),
  Xlab = NULL,
  ylab = NULL,
  Xlim = NULL,
  ylim = NULL,
  title = NULL,
  title_sep = " | ",
  add = FALSE,
  engine3d = NULL,
  ...
)

\method{sectionview3d}{WarpKriging}(
  WarpKriging_model,
  center = NULL,
  axis = NULL,
  npoints = 21,
  col_points = if (!is.null(col)) col else "red",
  col_fun = if (!is.null(col)) col else "blue",
  col = NULL,
  conf_level = 0.95,
  conf_fading = 0.5,
  bg_fading = 1,
  mfrow = c(1, 1),
  Xlab = NULL,
  ylab = NULL,
  Xlim = NULL,
  ylim = NULL,
  title = NULL,
  title_sep = " | ",
  add = FALSE,
  engine3d = NULL,
  ...
)

\method{sectionview3d}{glm}(
  glm_model,
  center = NULL,
  axis = NULL,
  npoints = 21,
  col_points = if (!is.null(col)) col else "red",
  col_fun = if (!is.null(col)) col else "blue",
  col = NULL,
  conf_level = 0.95,
  conf_fading = 0.5,
  bg_fading = 1,
  mfrow = c(1, 1),
  Xlab = NULL,
  ylab = NULL,
  Xlim = NULL,
  ylim = NULL,
  title = NULL,
  title_sep = " | ",
  add = FALSE,
  engine3d = NULL,
  ...
)

\method{sectionview3d}{list}(
  modelFit_model,
  center = NULL,
  axis = NULL,
  npoints = 21,
  col_points = if (!is.null(col)) col else "red",
  col_fun = if (!is.null(col)) col else "blue",
  col = NULL,
  bg_fading = 1,
  mfrow = c(1, 1),
  Xlab = NULL,
  ylab = NULL,
  Xlim = NULL,
  ylim = NULL,
  title = NULL,
  title_sep = " | ",
  add = FALSE,
  engine3d = NULL,
  ...
)

sectionview3d(...)
}
\arguments{
\item{fun}{a function or 'predict()'-like function that returns a simple numeric, or an interval, or mean and standard error: list(mean=...,se=...).}

\item{vectorized}{is fun vectorized?}

\item{center}{optional coordinates (as a list or data frame) of the center of the section view if the model's dimension is > 2.}

\item{axis}{optional matrix of 2-axis combinations to plot, one by row. The value \code{NULL} leads to all possible combinations i.e. \code{choose(D, 2)}.}

\item{npoints}{an optional number of points to discretize plot of response  surface and uncertainties.}

\item{col_fun}{color of the function plot.}

\item{col}{color of the object (use col_* for specific objects).}

\item{col_fading_interval}{an optional factor of alpha (color channel) fading used to plot confidence intervals.}

\item{mfrow}{an optional list to force \code{par(mfrow = ...)} call. The default value  \code{NULL} is automatically set for compact view.}

\item{Xlab}{an optional list of string to overload names for X.}

\item{ylab}{an optional string to overload name for y.}

\item{Xlim}{an optional list to force x range for all plots. The default value \code{NULL} is automatically set to include all design points (and their 1-99 percentiles).}

\item{ylim}{an optional list to force y range for all plots. The default value \code{NULL} is automatically set to include all design points (and their 1-99 percentiles).}

\item{title}{an optional overload of main title.}

\item{title_sep}{customize subtitle with fixed input.}

\item{add}{to print graphics on an existing window.}

\item{engine3d}{3D view package to use. "rgl" if available, otherwise "scatterplot3d" by default.}

\item{...}{arguments of the \code{sectionview3d.km}, \code{sectionview3d.glm}, \code{sectionview3d.Kriging} or \code{sectionview3d.function} function}

\item{X}{the matrix of input design.}

\item{y}{the array of output values (two columns means an interval).}

\item{col_points}{color of points.}

\item{bg_fading}{an optional factor of alpha (color channel) fading used to plot design points outside from this section.}

\item{eval_str}{the expression to evaluate in each subplot.}

\item{km_model}{an object of class \code{"km"}.}

\item{type}{the kriging type to use for model prediction.}

\item{conf_level}{an optional list of confidence intervals to display.}

\item{conf_fading}{an optional factor of alpha (color channel) fading used to plot confidence intervals.}

\item{Kriging_model}{an object of class \code{"Kriging"}.}

\item{WarpKriging_model}{an object of class \code{"WarpKriging"}.}

\item{glm_model}{an object of class \code{"glm"}.}

\item{modelFit_model}{an object returned by DiceEval::modelFit.}
}
\description{
Plot a contour view of a prediction model or function, including design points if available.
}
\details{
If available, experimental points are plotted with fading colors. Points that fall in the specified section (if any) have the color specified \code{col_points} while points far away from the center have shaded versions of the same color. The amount of fading is determined using the Euclidean distance between the plotted point and \code{center}.
}
\examples{
x1 <- rnorm(15)
x2 <- rnorm(15)

y <- x1 + x2 + rnorm(15)
model <- lm(y ~ x1 + x2)

sectionview3d(function(x) sum(x),
                       Xlim=cbind(range(x1),range(x2)), col='black')
DiceView:::.plot3d(x1, x2, y)

sectionview3d(function(x) {
                      x = as.data.frame(x)
                      colnames(x) <- all.vars(model$call)[-1]
                      p = predict.lm(model, newdata=x, se.fit=TRUE)
                      list(mean=p$fit, se=p$se.fit)
                    }, vectorized=TRUE,
              add=TRUE)

X = matrix(runif(15*2),ncol=2)
y = apply(X,1,branin)

sectionview3d(X, y)

x1 <- rnorm(15)
x2 <- rnorm(15)

y <- x1 + x2^2 + rnorm(15)
model <- glm(y ~ x1 + I(x2^2))

sectionview3d(model)

sectionview3d("abline(h=0.25,col='red')")
if (requireNamespace("DiceKriging")) { library(DiceKriging)

X = matrix(runif(15*2),ncol=2)
y = apply(X,1,branin)

model <- km(design = X, response = y, covtype="matern3_2")

sectionview3d(model)

}

if (requireNamespace("rlibkriging")) { library(rlibkriging)

X = matrix(runif(15*2),ncol=2)
y = apply(X,1,branin)

model <- Kriging(X = X, y = y, kernel="matern3_2")

sectionview3d(model)

}

if (requireNamespace("rlibkriging")) { library(rlibkriging)

X = matrix(runif(15*2),ncol=2)
y = apply(X,1,branin) + 5*rnorm(15)

model <- WarpKriging(y = y, X = X, warping = c("affine","affine"), kernel="matern3_2")

sectionview3d(model)

}

x1 <- rnorm(15)
x2 <- rnorm(15)

y <- x1 + x2^2 + rnorm(15)
model <- glm(y ~ x1 + I(x2^2))

sectionview3d(model)

if (requireNamespace("DiceEval")) { library(DiceEval)

X = matrix(runif(15*2),ncol=2)
y = apply(X,1,branin)

model <- modelFit(X, y, type = "StepLinear")

sectionview3d(model)

}

## A 2D example - Branin-Hoo function
sectionview3d(branin, col='black')

\dontrun{
## a 16-points factorial design, and the corresponding response
d <- 2; n <- 16
design.fact <- expand.grid(seq(0, 1, length = 4), seq(0, 1, length = 4))
design.fact <- data.frame(design.fact); names(design.fact) <- c("x1", "x2")
y <- branin(design.fact); names(y) <- "y"

if (requireNamespace("DiceKriging")) { library(DiceKriging)
## model: km
model <- DiceKriging::km(design = design.fact, response = y)
sectionview3d(model)
sectionview3d(branin, col='red', add=TRUE)
}

if (requireNamespace("rlibkriging")) { library(rlibkriging)
## model: Kriging
model <- Kriging(X = as.matrix(design.fact), y = as.matrix(y), kernel="matern3_2")
sectionview3d(model)
sectionview3d(branin, col='red', add=TRUE)
}

## model: glm
model <- glm(y ~ 1+ x1 + x2 + I(x1^2) + I(x2^2) + x1*x2, data=cbind(y,design.fact))
sectionview3d(model)
sectionview3d(branin, col='red', add=TRUE)

if (requireNamespace("DiceEval")) { library(DiceEval)
## model: StepLinear
model <- modelFit(design.fact, y, type = "StepLinear")
sectionview3d(model)
sectionview3d(branin, col='red', add=TRUE)
}
}

}
\seealso{
\code{\link{sectionview.function}} for a section plot, and \code{\link{sectionview3d.function}} for a 2D section plot.

\code{\link{sectionview.matrix}} for a section plot, and \code{\link{sectionview3d.matrix}} for a 2D section plot.

\code{\link{sectionview3d.matrix}} for a section plot.

\code{\link{sectionview.km}} for a section plot, and \code{\link{sectionview3d.km}} for a 2D section plot.

\code{\link{sectionview.Kriging}} for a section plot, and \code{\link{sectionview3d.Kriging}} for a 2D section plot.

\code{\link{sectionview.WarpKriging}} for a section plot, and \code{\link{sectionview3d.WarpKriging}} for a 2D section plot.

\code{\link{sectionview.glm}} for a section plot, and \code{\link{sectionview3d.glm}} for a 2D section plot.

\code{\link{sectionview.glm}} for a section plot, and \code{\link{sectionview3d.glm}} for a 2D section plot.
}
\author{
Yann Richet, ASNR
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/optim.R
\name{optims}
\alias{optims}
\title{Title Multi-local optimization wrapper for optim, using (possibly parallel) multistart.}
\usage{
optims(
  pars,
  fn,
  fn.NaN = NaN,
  fn.stop = NA,
  .apply = "mclapply",
  pars.eps = 1e-05,
  control = list(),
  ...
)
}
\arguments{
\item{pars}{starting points for optim}

\item{fn}{objective function, like in optim().}

\item{fn.NaN}{replacement value of fn when returns NaN}

\item{fn.stop}{early stopping criterion}

\item{.apply}{loop/parallelization backend for multistart ("mclapply", "lapply" or "foreach")}

\item{pars.eps}{minimal distance between two solutions to be considered different}

\item{control}{control parameters for optim()}

\item{...}{additional arguments passed to optim()}
}
\value{
list with best solution and all solutions
}
\description{
Title Multi-local optimization wrapper for optim, using (possibly parallel) multistart.
}
\examples{
fn = function(x) ifelse(x==0,1,sin(x)/x)
# plot(fn, xlim=c(-20,20))
optim( par=5,                     fn,lower=-20,upper=20,method='L-BFGS-B')
optims(pars=t(t(seq(-20,20,,20))),fn,lower=-20,upper=20,method='L-BFGS-B')

# Branin function (3 local minimas)
f = function (x) {
  x1 <- x[1] * 15 - 5
  x2 <- x[2] * 15
  (x2 - 5/(4 * pi^2) * (x1^2) + 5/pi * x1 - 6)^2 + 10 * (1 - 1/(8 * pi)) * cos(x1) + 10
}
# expect to find 3 local minimas
optims(pars=matrix(runif(100),ncol=2),f,method="L-BFGS-B",lower=c(0,0),upper=c(1,1))
}
\author{
Yann Richet, ASNR
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/mesh_set.R
\name{points_out.mesh}
\alias{points_out.mesh}
\title{Extract points of mesh which do not belong to the mesh triangulation (may not contain all points)}
\usage{
points_out.mesh(mesh)
}
\arguments{
\item{mesh}{(list(p,tri,...) from geometry)}
}
\value{
points coordinates outside the mesh triangulation
}
\description{
Extract points of mesh which do not belong to the mesh triangulation (may not contain all points)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/mesh_set.R
\name{mesh}
\alias{mesh}
\title{Builds a mesh from a design aor set of points}
\usage{
mesh(intervals, mesh.type = "seq", mesh.sizes = 11)
}
\arguments{
\item{intervals}{bounds to inverse in, each column contains min and max (or values) of each dimension}

\item{mesh.type}{function or "unif" or "seq" (default) or "LHS" to preform interval partition}

\item{mesh.sizes}{number of parts for mesh (duplicate for each dimension if using "seq")}
}
\value{
delaunay mesh (list(p,tri,...) from geometry)
}
\description{
Builds a mesh from a design aor set of points
}
\examples{
mesh = mesh(intervals=matrix(c(0,1,0,1),ncol=2),mesh.type="unif",mesh.sizes=10)
plot2d_mesh(mesh)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/root.R
\name{min_dist}
\alias{min_dist}
\title{Minimal distance between one point to many points}
\usage{
min_dist(x, X, norm = rep(1, ncol(X)))
}
\arguments{
\item{x}{one point}

\item{X}{matrix of points (same number of columns than x)}

\item{norm}{normalization vecor of distance (same number of columns than x)}
}
\value{
minimal distance
}
\description{
Minimal distance between one point to many points
}
\examples{
min_dist(runif(3),matrix(runif(30),ncol=3))
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/root.R
\name{mesh_level}
\alias{mesh_level}
\title{Mesh level set of function}
\usage{
mesh_level(f, vectorized = FALSE, level = 0, intervals, mesh, ...)
}
\arguments{
\item{f}{function to be evaluated on the mesh}

\item{vectorized}{logical or function. If TRUE, f is assumed to be vectorized.}

\item{level}{level/threshold value}

\item{intervals}{matrix of intervals}

\item{mesh}{mesh object or type}

\item{...}{additional arguments passed to f}
}
\description{
Mesh level set of function
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/mesh_set.R
\name{plot2d_mesh}
\alias{plot2d_mesh}
\title{Plot a two dimensional mesh}
\usage{
plot2d_mesh(
  mesh,
  color.nodes = "black",
  color.mesh = "darkgray",
  alpha = 0.4,
  ...
)
}
\arguments{
\item{mesh}{2-dimensional mesh to draw}

\item{color.nodes}{color of the mesh nodes}

\item{color.mesh}{color of the mesh elements}

\item{alpha}{transparency of the mesh elements & nodes}

\item{...}{optional arguments passed to plot function}
}
\description{
Plot a two dimensional mesh
}
\examples{
plot2d_mesh(mesh_exsets(f = function(x) sin(pi*x[1])*sin(pi*x[2]),
                        threshold=0,sign=1, mesh.type="unif",mesh.size=11,
                        intervals = matrix(c(1/2,5/2,1/2,5/2),nrow=2),
                        num_workers=1))
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/mesh_set.R
\name{is_in.mesh}
\alias{is_in.mesh}
\title{Checks if some point belongs to a given mesh}
\usage{
is_in.mesh(x, mesh)
}
\arguments{
\item{x}{point to check}

\item{mesh}{mesh identifying the set which X may belong}
}
\description{
Checks if some point belongs to a given mesh
}
\examples{
is_in.mesh(-0.5,mesh=geometry::delaunayn(matrix(c(0,1),ncol=1),output.options =TRUE))
is_in.mesh(0.5,mesh=geometry::delaunayn(matrix(c(0,1),ncol=1),output.options =TRUE))

x =matrix(-.5,ncol=2,nrow=1)
is_in.mesh(x,mesh=geometry::delaunayn(matrix(c(0,0,1,1,0,0),ncol=2),output.options =TRUE))

x =matrix(.5,ncol=2,nrow=1)
is_in.mesh(x,mesh=geometry::delaunayn(matrix(c(0,0,1,1,0,0),ncol=2),output.options =TRUE))
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/mesh_set.R
\name{mesh_exsets}
\alias{mesh_exsets}
\title{Search excursion set of nD function, sampled by a mesh}
\usage{
mesh_exsets(
  f,
  vectorized = FALSE,
  threshold,
  sign,
  intervals,
  mesh.type = "seq",
  mesh.sizes = 11,
  maxerror_f = 1e-09,
  tol = .Machine$double.eps^0.25,
  ex_filter.tri = all,
  num_workers = maxWorkers(),
  ...
)
}
\arguments{
\item{f}{Function to inverse at 'threshold'}

\item{vectorized}{boolean: is f already vectorized ? (default: FALSE) or if function: vectorized version of f.}

\item{threshold}{target value to inverse}

\item{sign}{focus at conservative for above (sign=1) or below (sign=-1) the threshold}

\item{intervals}{bounds to inverse in, each column contains min and max of each dimension}

\item{mesh.type}{"unif" or "seq" (default) or "LHS" to preform interval partition}

\item{mesh.sizes}{number of parts for mesh (duplicate for each dimension if using "seq")}

\item{maxerror_f}{maximal tolerance on f precision}

\item{tol}{the desired accuracy (convergence tolerance on f arg).}

\item{ex_filter.tri}{boolean function to validate a geometry::tri as considered in excursion : 'any' or 'all'}

\item{num_workers}{number of cores to use for parallelization}

\item{...}{parameters to forward to roots_mesh(...) call}
}
\description{
Search excursion set of nD function, sampled by a mesh
}
\examples{
# mesh_exsets(function(x) x, threshold=.51, sign=1, intervals=rbind(0,1),
#   maxerror_f=1E-2,tol=1E-2, num_workers=1) # for faster testing
# mesh_exsets(function(x) x, threshold=.50000001, sign=1, intervals=rbind(0,1),
#   maxerror_f=1E-2,tol=1E-2, num_workers=1) # for faster testing
# mesh_exsets(function(x) sum(x), threshold=.51,sign=1, intervals=cbind(rbind(0,1),rbind(0,1)),
#   maxerror_f=1E-2,tol=1E-2, num_workers=1) # for faster testing
# mesh_exsets(sin,threshold=0,sign="sup",interval=c(pi/2,5*pi/2),
#   maxerror_f=1E-2,tol=1E-2, num_workers=1) # for faster testing

if (identical(Sys.getenv("NOT_CRAN"), "true")) { # too long for CRAN on Windows

  e = mesh_exsets(function(x) (0.25+x[1])^2+(0.5+x[2])^2 ,
                threshold =0.25,sign=-1, intervals=matrix(c(-1,1,-1,1),nrow=2),
                maxerror_f=1E-2,tol=1E-2, # for faster testing
                num_workers=1)

  plot(e$p,xlim=c(-1,1),ylim=c(-1,1));
  apply(e$tri,1,function(tri) polygon(e$p[tri,],col=rgb(.4,.4,.4,.4)))
  apply(e$frontiers,1,function(front) lines(e$p[front,],col='red'))

  if (requireNamespace("rgl")) {
    e = mesh_exsets(function(x) (0.5+x[1])^2+(-0.5+x[2])^2+(0.+x[3])^2,
                  threshold = .25,sign=-1, mesh.type="unif",
                  intervals=matrix(c(-1,1,-1,1,-1,1),nrow=2),
                  maxerror_f=1E-2,tol=1E-2, # for faster testing
                  num_workers=1)

    rgl::plot3d(e$p,xlim=c(-1,1),ylim=c(-1,1),zlim=c(-1,1));
    apply(e$tri,1,function(tri)rgl::lines3d(e$p[tri,]))
  }
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/Utils.R
\name{branin}
\alias{branin}
\title{This is a simple copy of the Branin-Hoo 2-dimensional test function, as provided in DiceKriging package.
The Branin-Hoo function is defined here over [0,1] x [0,1], instead of [-5,0] x [10,15] as usual.
It has 3 global minima : x1 = c(0.9616520, 0.15); x2 = c(0.1238946, 0.8166644); x3 = c(0.5427730, 0.15)}
\usage{
branin(x)
}
\arguments{
\item{x}{a 2-dimensional vector specifying the location where the function is to be evaluated.}
}
\value{
A real number equal to the Branin-Hoo function values at x
}
\description{
This is a simple copy of the Branin-Hoo 2-dimensional test function, as provided in DiceKriging package.
The Branin-Hoo function is defined here over [0,1] x [0,1], instead of [-5,0] x [10,15] as usual.
It has 3 global minima : x1 = c(0.9616520, 0.15); x2 = c(0.1238946, 0.8166644); x3 = c(0.5427730, 0.15)
}
